
from __future__ import annotations

import numpy as np
import torch
from typing import List, Union, Type
from copy import deepcopy
from pyjuice.utils import BitSet


class RegionGraph():

    # Property checks
    ALLOW_NONSMOOTH = False
    ALLOW_NONDECOMPOSABLE = False

    def __init__(self, scope: BitSet, children: List[RegionGraph]) -> None:
        self.scope = scope
        self.children = children
        self.num_chs = len(children)

        self.parents = []
        for region_node in self.children:
            region_node.parents.append(self)

    
class PartitionNode(RegionGraph):
    def __init__(self, children: List[Union[InnerRegionNode,InputRegionNode]]) -> None:

        assert len(children) > 0, "PartitionNode receives no child."

        scope = BitSet()
        for n in children:
            if not self.ALLOW_NONDECOMPOSABLE:
                assert len(scope & n.scope) == 0, "Children of a PartitionNode have overlapping scopes."
            scope |= n.scope

        super().__init__(scope, children)

    def __hash__(self):
        ch_scopes = tuple(hash(c.scope) for c in self.children)
        return hash(("PartitionNode", hash(self.scope), ch_scopes))


class InnerRegionNode(RegionGraph):
    def __init__(self, children: List[Union[InputRegionNode,PartitionNode]]) -> None:

        assert len(children) > 0, "InnerRegionNode receives no child."

        scope = deepcopy(children[0].scope)
        for n in children[1:]:
            if not self.ALLOW_NONSMOOTH:
                assert scope == n.scope, "Children of an InnerRegionNode must have the same scope."
            else:
                scope |= n.scope

        super().__init__(scope, children)

    def __hash__(self):
        return hash(("InnerRegionNode", hash(self.scope)))


class InputRegionNode(RegionGraph):
    def __init__(self, scope: Union[List,BitSet]) -> None:
        if isinstance(scope, List):
            scope = BitSet.from_array(scope)

        super().__init__(scope, [])

    def __hash__(self):
        return hash(("InputRegionNode", hash(self.scope)))



from __future__ import annotations

import numpy as np
import torch
from typing import Sequence, Union, Type
from copy import deepcopy
from functools import reduce

from pyjuice.graph import InnerRegionNode
from .backend import normalize_ns_parameters
from .nodes import CircuitNodes
from .prod_nodes import ProdNodes

Tensor = Union[np.ndarray,torch.Tensor]


class SumNodes(CircuitNodes):
    """
    A class representing vectors of sum nodes. It is created by :code:`pyjuice.summate`.

    :param num_node_blocks: number of node blocks
    :type num_node_blocks: int

    :param chs: sequence of child nodes
    :type chs: Sequence[CircuitNodes]

    :param edge_ids: a matrix of size [2, # edges] - every size-2 column vector [i,j] defines a set of edges that fully connect the ith sum node block and the jth child node block
    :type edge_ids: Optional[Tensor]

    :param block_size: block size
    :type block_size: int
    """

    def __init__(self, num_node_blocks: int, chs: Sequence[CircuitNodes], edge_ids: Optional[Union[Tensor,Sequence[Tensor]]] = None, 
                 params: Optional[Tensor] = None, zero_param_mask: Optional[Tensor] = None, block_size: int = 0, **kwargs) -> None:

        assert len(chs) > 0, "`SumNodes` must have at least one child."
        for i in range(1, len(chs)):
            assert chs[0].block_size == chs[i].block_size, "`block_size` of the children of a `SumNodes` should be the same."

        rg_node = InnerRegionNode([ch.region_node for ch in chs])
        super(SumNodes, self).__init__(num_node_blocks, rg_node, block_size = block_size, **kwargs)

        # Child layers
        self.chs = self._standardize_chs(chs)

        # Total number of child circuit node blocks
        self.num_ch_node_blocks = reduce(lambda m, n: m + n, map(lambda n: n.num_node_blocks, chs))

        # Block size of the children
        self.ch_block_size = self.chs[0].block_size

        # Construct sum edges
        self._construct_edges(edge_ids)

        # Set zero parameter mask
        if zero_param_mask is not None:
            self.set_zero_param_mask(zero_param_mask)

        # Set parameters
        if params is not None:
            self.set_params(params, pseudocount = 0.0)

        # Callbacks
        self._run_init_callbacks(**kwargs)

    @property
    def num_edges(self):
        """
        Number of edges within the current node.
        """
        return self.edge_ids.size(1) * self.block_size * self.ch_block_size

    @property
    def num_ch_nodes(self):
        return self.num_ch_node_blocks * self.ch_block_size

    def duplicate(self, *args, tie_params: bool = False) -> SumNodes:
        """
        Create a duplication of the current node with the same specification (i.e., number of nodes, block size).

        :note: The child nodes should have the same specifications compared to the original child nodes.

        :param args: a sequence of new child nodes
        :type args: CircuitNodes

        :param tie_params: whether to tie the parameters of the current node and the duplicated node
        :type tie_params: bool

        :returns: a duplicated :code:`SumNodes`
        """
        chs = []
        for ns in args:
            assert isinstance(ns, CircuitNodes)
            chs.append(ns)

        if len(chs) == 0:
            chs = self.chs
        else:
            assert self.num_chs == len(chs), f"Number of new children ({len(chs)}) must match the number of original children ({self.num_chs})."
            for old_c, new_c in zip(self.chs, chs):
                assert type(old_c) == type(new_c), f"Child type not match: ({type(new_c)} != {type(old_c)})."
                assert old_c.num_node_blocks == new_c.num_node_blocks, f"Child node size not match: (`num_node_blocks`: {new_c.num_node_blocks} != {old_c.num_node_blocks})."
                assert old_c.block_size == new_c.block_size, f"Child node size not match: (`block_size`: {new_c.block_size} != {old_c.block_size})."

        edge_ids = self.edge_ids.clone()

        if hasattr(self, "_params") and self._params is not None and not tie_params:
            params = self._params.clone()
        else:
            # We also do not copy parameters explicitly if this is a tied node
            params = None

        return SumNodes(self.num_node_blocks, chs, edge_ids, params = params, block_size = self.block_size, source_node = self if tie_params else None)

    def get_params(self):
        """
        Get the sum node parameters.
        """
        if not hasattr(self, "_params"):
            return None
        return self._params

    def set_params(self, params: torch.Tensor, normalize: bool = True, pseudocount: float = 0.0):
        """
        Set the sum node parameters.

        :param params: parameters to be set
        :type params: Union[torch.Tensor,Dict]

        :param normalize: whether to normalize the parameters
        :type normalize: bool

        :param pseudocount: pseudo count added to the parameters
        :type pseudocount: float
        """
        if self._source_node is not None:
            ns_source = self._source_node
            ns_source.set_params(params, normalize = normalize, pseudocount = pseudocount)

            return None

        if params.dim() == 1:
            assert self.block_size == 1 and self.ch_block_size == 1
            assert self.edge_ids.size(1) == params.size(0)

            self._params = params.clone().view(-1, 1, 1)

        elif params.dim() == 2:
            ch_num_nblocks = sum([cs.num_node_blocks for cs in self.chs])
            assert params.size(0) == self.num_nodes
            assert params.size(1) == self.ch_block_size * ch_num_nblocks

            self._params = params.reshape(self.num_node_blocks, self.block_size, ch_num_nblocks, self.ch_block_size).permute(0, 2, 1, 3).flatten(0, 1).contiguous()

        elif params.dim() == 3:
            assert self.edge_ids.size(1) == params.size(0) and self.block_size == params.size(1) and self.ch_block_size == params.size(2)

            self._params = params.clone()

        elif params.dim() == 4:
            assert params.size(0) == self.num_node_blocks and params.size(1) == self.num_ch_node_blocks and \
                self.block_size == params.size(2) and self.ch_block_size == params.size(3)

            self._params = params[self.edge_ids[0,:],self.edge_ids[1,:],:,:].clone().contiguous()

        else:
            raise ValueError("Unsupported parameter input.")

        if self.provided("zero_param_mask"):
            self._params[self._zero_param_mask] = 0.0

        if normalize:
            normalize_ns_parameters(self._params, self.edge_ids[0,:], block_size = self.block_size, 
                                    ch_block_size = self.ch_block_size, pseudocount = pseudocount)

    def set_zero_param_mask(self, zero_param_mask: Optional[Tensor] = None):
        if zero_param_mask is None:
            return None

        if self._source_node is not None:
            ns_source = self._source_node
            ns_source.set_zero_param_mask(zero_param_mask)

            return None

        assert zero_param_mask.dim() == 3
        assert zero_param_mask.size(0) == self.edge_ids.size(1)
        assert zero_param_mask.size(1) == self.block_size
        assert zero_param_mask.size(2) == self.ch_block_size
        assert zero_param_mask.dtype == torch.bool

        self._zero_param_mask = zero_param_mask

    def get_zero_param_mask(self):
        if not self.provided("_zero_param_mask"):
            return None
        else:
            return self._zero_param_mask

    def set_edges(self, edge_ids: Union[Tensor,Sequence[Tensor]]):
        self._construct_edges(edge_ids)

        self._params = None # Clear parameters

    def init_parameters(self, perturbation: float = 2.0, recursive: bool = True, is_root: bool = True, **kwargs):
        """
        Randomly initialize node parameters.

        :param perturbation: "amount of perturbation" added to the parameters (should be greater than 0)
        :type perturbation: float

        :param recursive: whether to recursively apply the function to child nodes
        :type recursive: bool
        """
        if self._source_node is None:
            self._params = torch.exp(torch.rand([self.edge_ids.size(1), self.block_size, self.ch_block_size]) * -perturbation)

            if self.provided("zero_param_mask"):
                self._params[self._zero_param_mask] = 0.0

            normalize_ns_parameters(self._params, self.edge_ids[0,:], block_size = self.block_size, 
                                    ch_block_size = self.ch_block_size, pseudocount = 0.0)

        super(SumNodes, self).init_parameters(
            perturbation = perturbation, 
            recursive = recursive, 
            is_root = is_root, 
            **kwargs
        )

    def update_parameters(self, params: torch.Tensor, clone: bool = True):
        """
        Update parameters from `pyjuice.TensorCircuit` to the current node.

        :param params: the parameter tensor in the `TensorCircuit`
        :type params: torch.Tensor

        :param clone: whether to clone the parameters
        :type clone: bool
        """
        assert self.provided("_param_range"), "The `SumNodes` has not been compiled into a `TensorCircuit`."

        if self.is_tied():
            # Do not update parameters for tied nodes
            return None

        psid, peid = self._param_range
        if clone:
            ns_params = params[psid:peid].cpu().clone()
        else:
            ns_params = params[psid:peid].cpu()

        local_parids = (self._param_ids - psid) // (self.block_size * self.ch_block_size)
        num_parblocks = local_parids.size(0)
        ns_params = ns_params.reshape(num_parblocks, self.ch_block_size, self.block_size)
        self._params = ns_params[local_parids,:,:].permute(0, 2, 1)

    def update_param_flows(self, param_flows: torch.Tensor, origin_ns_only: bool = True, clone: bool = True):
        """
        Update parameter flows from `pyjuice.TensorCircuit` to the current node.

        :param params_flows: the parameter flow tensor in the `TensorCircuit`
        :type params_flows: torch.Tensor

        :param clone: whether to clone the parameters
        :type clone: bool
        """
        assert self.provided("_param_flow_range"), "The `SumNodes` has not been compiled into a `TensorCircuit`."

        if origin_ns_only and self.is_tied():
            return None

        pfsid, pfeid = self._param_flow_range
        if clone:
            ns_param_flows = param_flows[pfsid:pfeid].cpu().clone()
        else:
            ns_param_flows = param_flows[pfsid:pfeid].cpu()

        local_parfids = (self._param_flow_ids - pfsid) // (self.block_size * self.ch_block_size)
        num_parfblocks = local_parfids.size(0)
        ns_param_flows = ns_param_flows.reshape(num_parfblocks, self.ch_block_size, self.block_size)
        self._param_flows = ns_param_flows[local_parfids,:,:].permute(0, 2, 1)

    def gather_parameters(self, params: torch.Tensor):
        """
        Update parameters from the current node to the compiled `pyjuice.TensorCircuit`.

        :param params: the parameter tensor in the `TensorCircuit`
        :type params: torch.Tensor
        """
        assert self.provided("_param_range"), "The `SumNodes` has not been compiled into a `TensorCircuit`."

        if self.is_tied() or not self.has_params():
            return None

        psid, peid = self._param_range
        ns_params = self._params[self._inverse_param_ids,:,:].permute(0, 2, 1).reshape(-1)
        params[psid:peid] = ns_params.to(params.device)

    def _get_edges_as_mask(self):
        mask = torch.zeros([self.num_node_blocks, self.num_ch_nodes], dtype = torch.bool)
        mask[self.edge_ids[0,:], self.edge_ids[1,:]] = True

        return mask

    def _standardize_chs(self, chs):
        new_chs = []
        for cs in chs:
            if cs.is_input():
                new_cs = ProdNodes(
                    num_node_blocks = cs.num_node_blocks,
                    chs = [cs],
                    edge_ids = torch.arange(0, cs.num_node_blocks).reshape(-1, 1),
                    block_size = cs.block_size
                )
                new_chs.append(new_cs)
            else:
                new_chs.append(cs)

        return new_chs

    def _construct_edges(self, edge_ids: Optional[Union[Tensor,Sequence[Tensor]]], reorder: bool = False):
        if edge_ids is None:
            edge_ids = torch.cat(
                (torch.arange(self.num_node_blocks).unsqueeze(1).repeat(1, self.num_ch_node_blocks).reshape(1, -1),
                 torch.arange(self.num_ch_node_blocks).unsqueeze(0).repeat(self.num_node_blocks, 1).reshape(1, -1)),
                dim = 0
            )
        elif isinstance(edge_ids, Sequence):
            assert len(edge_ids) == len(self.chs)

            per_ns_edge_ids = edge_ids
            ch_gid_start = 0
            edge_ids = []
            for cs_id in range(len(self.chs)):
                curr_edge_ids = per_ns_edge_ids[cs_id]
                curr_edge_ids[1,:] += ch_gid_start
                edge_ids.append(curr_edge_ids)

            edge_ids = torch.cat(edge_ids, dim = 1)

        if reorder:
            edge_ids = self._reorder_edges(edge_ids)

        if isinstance(edge_ids, np.ndarray):
            edge_ids = torch.from_numpy(edge_ids)

        if edge_ids.dim() == 2 and edge_ids.type() == torch.bool:
            assert edge_ids.size(0) == self.num_node_blocks and edge_ids.size(1) == self.num_ch_node_blocks
            x_ids, y_ids = torch.where(edge_ids)
            edge_ids = torch.stack((x_ids, y_ids), dim = 0)

        # Sanity checks
        assert edge_ids.size(0) == 2, "Expect `edge_ids.size(0) == 2`."
        assert torch.all(edge_ids[0,:] >= 0) and torch.all(edge_ids[1,:] >= 0), "Edge index underflow."
        assert torch.all(edge_ids[0,:] < self.num_node_blocks) and torch.all(edge_ids[1,:] < self.num_ch_node_blocks), "Edge index overflow."
        par_ns = torch.unique(edge_ids[0,:])
        assert par_ns.size(0) == self.num_node_blocks and par_ns.max() == self.num_node_blocks - 1, "Some node has no edge."

        self.edge_ids = edge_ids

    def _reorder_edges(self, edge_ids: Tensor):
        ids = torch.argsort(edge_ids[0,:] * self.num_ch_node_blocks + edge_ids[1,:])
        return edge_ids[:,ids].contiguous()

    def __repr__(self):
        scope_size = len(self.scope)
        return f"SumNodes(num_node_blocks={self.num_node_blocks}, block_size={self.block_size}, num_chs={self.num_chs}, num_edges={self.num_edges}, scope_size={scope_size})"



from __future__ import annotations

import torch
import triton
import triton.language as tl

from typing import Optional, Any

from .distributions import Distribution


class Bernoulli(Distribution):
    """
    A class representing Bernoulli distributions.
    """
    def __init__(self):
        super(Bernoulli, self).__init__()

    def get_signature(self):
        """
        Get the signature of the current distribution.
        """
        return "Bernoulli"

    def get_metadata(self):
        """
        Get the metadata of the current distribution.
        """
        return []

    def num_parameters(self):
        """
        The number of parameters per node.
        """
        return 1

    def num_param_flows(self):
        """
        The number of parameter flows per node.
        """
        return 2

    def init_parameters(self, num_nodes: int, perturbation: float = 2.0, params: Optional[Any] = None, **kwargs):
        """
        Initialize parameters for `num_nodes` nodes.
        Returned parameters should be flattened into a vector.
        """
        if params is not None:
            assert isinstance(params, torch.Tensor)
            assert params.numel() == num_nodes * self.num_parameters()
            return params

        params = torch.exp(torch.rand([num_nodes, 2]) * -perturbation)
        params /= params.sum(dim = 1, keepdim = True)

        return params[:,0].contiguous()

    def get_data_dtype(self):
        """
        Get the data dtype for the distribution.
        """
        return torch.bool

    @staticmethod
    def fw_mar_fn(local_offsets, data, params_ptr, s_pids, metadata_ptr, s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE):
        probs = tl.load(params_ptr + s_pids, mask = mask, other = 0)
        probs = tl.where(data > 0, probs, 1.0 - probs)
        log_probs = tl.log(probs)

        return log_probs

    @staticmethod
    def bk_flow_fn(local_offsets, ns_offsets, data, flows, node_mars_ptr, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, 
                   s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE):
        pf_offsets = s_pfids + tl.where(data > 0, 0, 1)
        tl.atomic_add(param_flows_ptr + pf_offsets, flows, mask = mask)

    @staticmethod
    def sample_fn(samples_ptr, local_offsets, batch_offsets, vids, s_pids, params_ptr, metadata_ptr, s_mids_ptr, mask, batch_size, BLOCK_SIZE, seed):

        rnd_val = tl.rand(seed, tl.arange(0, BLOCK_SIZE))

        param = tl.load(params_ptr + s_pids, mask = mask, other = 0)
        sampled_id = tl.where(rnd_val < param, 1, 0)

        # Write back to `samples`
        sample_offsets = vids * batch_size + batch_offsets
        tl.store(samples_ptr + sample_offsets, sampled_id, mask = mask)

    @staticmethod
    def em_fn(local_offsets, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, s_mids_ptr, mask,
              step_size, pseudocount, BLOCK_SIZE):

        flow1 = tl.load(param_flows_ptr + s_pfids, mask = mask, other = 0)
        flow0 = tl.load(param_flows_ptr + s_pfids + 1, mask = mask, other = 0)

        param = tl.load(params_ptr + s_pids, mask = mask, other = 0)
        updated_param = (flow1 + 0.5 * pseudocount) / (flow1 + flow0 + pseudocount)

        new_param = (1.0 - step_size) * param + step_size * updated_param
        tl.store(params_ptr + s_pids, new_param, mask = mask)

    def _get_constructor(self):
        return Bernoulli, {}


from __future__ import annotations

import numpy as np
import torch
from typing import Sequence, Union, Type, Optional
from copy import deepcopy
from functools import reduce

from pyjuice.graph import PartitionNode
from .nodes import CircuitNodes

Tensor = Union[np.ndarray,torch.Tensor]


class ProdNodes(CircuitNodes):
    """
    A class representing vectors of product nodes. It is created by :code:`pyjuice.multiply`.

    :param num_node_blocks: number of node blocks
    :type num_node_blocks: int

    :param chs: sequence of child nodes
    :type chs: Sequence[CircuitNodes]

    :param edge_ids: a matrix of size [# product node blocks, # children] - the ith product node block is connected to the `edge_ids[i,j]`th node block in the jth child
    :type edge_ids: Optional[Tensor]

    :param block_size: block size
    :type block_size: int
    """

    SPARSE = 0
    BLOCK_SPARSE = 1

    def __init__(self, num_node_blocks: int, chs: Sequence[CircuitNodes], edge_ids: Optional[Tensor] = None, block_size: int = 0, **kwargs) -> None:

        rg_node = PartitionNode([ch.region_node for ch in chs])
        super(ProdNodes, self).__init__(num_node_blocks, rg_node, block_size = block_size, **kwargs)

        # Child layers
        self.chs = chs

        # Construct product edges
        self._construct_edges(edge_ids)

        # Callbacks
        self._run_init_callbacks(**kwargs)

    @property
    def num_edges(self):
        """
        Number of edges within the current node.
        """
        return self.num_nodes * self.num_chs

    @property
    def edge_type(self):
        """
        Type of the product edge. Either :code:`BLOCK_SPARSE` or :code:`SPARSE`.
        """
        if self.edge_ids.size(0) == self.num_node_blocks:
            return self.BLOCK_SPARSE
        elif self.edge_ids.size(0) == self.num_nodes:
            return self.SPARSE
        else:
            raise RuntimeError(f"Unexpected shape of `edge_ids`: ({self.edge_ids.size(0)}, {self.edge_ids.size(1)})")

    def is_block_sparse(self):
        """
        Whether the edge type is :code:`BLOCK_SPARSE`.
        """
        return self.edge_type == self.BLOCK_SPARSE

    def is_sparse(self):
        """
        Whether the edge type is :code:`SPARSE`.
        """
        return self.edge_type == self.SPARSE

    def duplicate(self, *args, tie_params: bool = False, allow_type_mismatch: bool = False):
        """
        Create a duplication of the current node with the same specification (i.e., number of nodes, block size).

        :note: The child nodes should have the same specifications compared to the original child nodes.

        :param args: a sequence of new child nodes
        :type args: CircuitNodes

        :param tie_params: whether to tie the parameters of the current node and the duplicated node
        :type tie_params: bool

        :returns: a duplicated :code:`ProdNodes`
        """
        chs = []
        for ns in args:
            assert isinstance(ns, CircuitNodes)
            chs.append(ns)

        if len(chs) == 0:
            chs = self.chs
        else:
            assert self.num_chs == len(chs), f"Number of new children ({len(chs)}) must match the number of original children ({self.num_chs})."
            for old_c, new_c in zip(self.chs, chs):
                if not allow_type_mismatch:
                    assert type(old_c) == type(new_c), f"Child type not match: ({type(new_c)} != {type(old_c)})."
                else:
                    assert not new_c.is_prod(), f"Cannot connect a product node to another."
                assert old_c.num_node_blocks == new_c.num_node_blocks, f"Child node size not match: (`num_node_blocks`: {new_c.num_node_blocks} != {old_c.num_node_blocks})."
                assert old_c.block_size == new_c.block_size, f"Child node size not match: (`block_size`: {new_c.block_size} != {old_c.block_size})."

        edge_ids = self.edge_ids.clone()

        return ProdNodes(self.num_node_blocks, chs, edge_ids, block_size = self.block_size, source_node = self if tie_params else None)

    def init_parameters(self, perturbation: float = 2.0, recursive: bool = True, is_root: bool = True, **kwargs):
        """
        Randomly initialize node parameters.

        :param perturbation: "amount of perturbation" added to the parameters (should be greater than 0)
        :type perturbation: float

        :param recursive: whether to recursively apply the function to child nodes
        :type recursive: bool
        """
        super(ProdNodes, self).init_parameters(
            perturbation = perturbation, 
            recursive = recursive, 
            is_root = is_root,
            **kwargs
        )

    def _construct_edges(self, edge_ids: Optional[Tensor]):
        if edge_ids is None:
            for c in self.chs:
                assert self.num_node_blocks == c.num_node_blocks and self.block_size == c.block_size, \
                    "Cannot create edges implicitly since # nodes do not match."

            edge_ids = torch.arange(self.num_node_blocks).unsqueeze(1).repeat(1, self.num_chs)

        if isinstance(edge_ids, np.ndarray):
            edge_ids = torch.from_numpy(edge_ids)

        # Sanity checks
        if edge_ids.size(0) == self.num_node_blocks:
            assert edge_ids.size(0) == self.num_node_blocks and edge_ids.size(1) == self.num_chs, f"Expect edge_ids.size() == ({self.num_node_blocks}, {self.num_chs})."
            for cid in range(self.num_chs):
                assert torch.all(edge_ids[:,cid] >= 0), "Edge index underflow."
                assert torch.all(edge_ids[:,cid] < self.chs[cid].num_node_blocks), "Edge index overflow."
        elif edge_ids.size(0) == self.num_nodes:
            assert edge_ids.size(0) == self.num_nodes and edge_ids.size(1) == self.num_chs, f"Expect edge_ids.size() == ({self.num_nodes}, {self.num_chs})."
            for cid in range(self.num_chs):
                assert torch.all(edge_ids[:,cid] >= 0), "Edge index underflow."
                assert torch.all(edge_ids[:,cid] < self.chs[cid].num_nodes), "Edge index overflow."
        else:
            raise RuntimeError(f"Unexpected shape of `edge_ids`: ({edge_ids.size(0)}, {edge_ids.size(1)})")

        self.edge_ids = edge_ids

    def __repr__(self):
        edge_type = "sparse" if self.edge_type == self.SPARSE else "block_sparse"
        scope_size = len(self.scope)
        return f"ProdNodes(num_node_blocks={self.num_node_blocks}, block_size={self.block_size}, num_chs={self.num_chs}, edge_type='{edge_type}', scope_size={scope_size})"



from __future__ import annotations

import torch
import numpy as np
from typing import Union, Sequence
from copy import deepcopy

from pyjuice.utils.context_manager import _DecoratorContextManager
from pyjuice.utils import BitSet
from .nodes import CircuitNodes
from .input_nodes import InputNodes
from .prod_nodes import ProdNodes
from .sum_nodes import SumNodes
from .distributions import Distribution
from pyjuice.graph import RegionGraph

Tensor = Union[np.ndarray,torch.Tensor]
ProdNodesChs = Union[SumNodes,InputNodes]
SumNodesChs = Union[ProdNodes,InputNodes]


def inputs(var: Union[int,Sequence[int]], num_node_blocks: int = 0, dist: Distribution = Distribution(), 
           params: Optional[Tensor] = None, num_nodes: int = 0, block_size: int = 0, **kwargs) -> InputNodes:
    """
    Construct a vector of input nodes defined on the same variable and have the same input distribution.

    :param var: the variable ID (or a list of variable IDs if the parameter `dist` is a multivariate distribution)
    :type var: Union[int,Sequence[int]]

    :param num_node_blocks: number of node blocks each of size `block_size`. I.e., the total number of nodes is `num_node_blocks` * `block_size`
    :type num_node_blocks: int

    :param dist: the input distribution from `pyjuice.nodes.distributions`
    :type dist: Distribution

    :param params: an optional tensor that defines the parameters of the nodes
    :type params: Optional[Tensor]

    :param num_nodes: an alternative of `num_node_blocks`; if set, it should be divicible by `block_size`
    :type num_nodes: int

    :param block_size: block size of the nodes; it does not change the semantics of the nodes, but will affect the speed of the compiled PC
    :type block_size: int

    :returns: an `InputNodes` object (a subclass of `CircuitNodes`)
    """

    assert block_size == 0 or block_size & (block_size - 1) == 0, "`block_size` must be a power of 2."

    if num_nodes > 0:
        assert num_node_blocks == 0, "Only one of `num_nodes` and `num_node_blocks` can be set at the same time."
        if block_size == 0:
            block_size = CircuitNodes.DEFAULT_BLOCK_SIZE

        assert num_nodes % block_size == 0

        num_node_blocks = num_nodes // block_size

    return InputNodes(
        num_node_blocks = num_node_blocks,
        scope = [var] if isinstance(var, int) else var,
        dist = dist,
        params = params,
        block_size = block_size,
        **kwargs
    )


def multiply(nodes1: ProdNodesChs, *args, edge_ids: Optional[Tensor] = None, sparse_edges: bool = False, **kwargs) -> ProdNodes:
    """
    Construct a vector of product nodes given a list of children PCs defined on disjoint sets of variables.

    :note: It requires all children to have the same block size.

    :note: If all children have the same `num_node_blocks`, if `edge_ids` is not specified, `multiply` outputs a vector of `num_node_blocks` * `block_size` nodes, where the ith (product) node is connected to the ith node in every child.

    :note: By default every edge specified by `edge_ids` denotes a block of `block_size` edges, unless `sparse_edge` is set to `True`, where the block size is assumed to be 1.

    :param nodes1: the first child node
    :type nodes1: Union[SumNodes,InputNodes]

    :param args: the remaining child nodes
    :type args: Union[SumNodes,InputNodes]

    :param edge_ids: a matrix of size [# product node blocks, # children] - the ith product node block is connected to the `edge_ids[i,j]`th node block in the jth child
    :type edge_ids: Optional[Tensor]

    :param sparse_edges: if set to `True`, the size of `edge_ids` becomes [# product nodes, # children] (i.e., we assume the block size of every child is 1)
    :type sparse_edges: bool

    :returns: an `ProdNodes` object (a subclass of `CircuitNodes`)
    """

    assert isinstance(nodes1, SumNodes) or isinstance(nodes1, InputNodes), "Children of product nodes must be input or sum nodes." 

    chs = [nodes1]
    num_node_blocks = nodes1.num_node_blocks
    block_size = nodes1.block_size
    scope = deepcopy(nodes1.scope)

    for nodes in args:
        assert nodes.is_input() or nodes.is_sum(), f"Children of product nodes must be input or sum nodes, but found input of type {type(nodes)}."
        if edge_ids is None:
            assert nodes.num_node_blocks == num_node_blocks, f"Input nodes should have the same `num_node_blocks`, but got {nodes.num_node_blocks} and {num_node_blocks}."
        assert nodes.block_size == block_size, "Input nodes should have the same `num_node_blocks`."
        if not RegionGraph.ALLOW_NONDECOMPOSABLE:
            assert len(nodes.scope & scope) == 0, "Children of a `ProdNodes` should have disjoint scopes."
        chs.append(nodes)
        scope |= nodes.scope

    if edge_ids is not None:
        if sparse_edges:
            assert edge_ids.shape[0] % block_size == 0
            num_node_blocks = edge_ids.shape[0] // block_size
        else:
            num_node_blocks = edge_ids.shape[0]
        assert edge_ids.shape[0] == num_node_blocks or edge_ids.shape[0] == num_node_blocks * block_size

    return ProdNodes(num_node_blocks, chs, edge_ids, block_size = block_size, **kwargs)


def summate(nodes1: SumNodesChs, *args, num_node_blocks: int = 0, num_nodes: int = 0,
            edge_ids: Optional[Tensor] = None, block_size: int = 0, **kwargs) -> SumNodes:
    """
    Construct a vector of sum nodes given a list of children PCs defined on the same sets of variables.

    :note: It requires all children to have the same block size.

    :note: If `edge_ids` is not set, `summate` defines a set of sum nodes fully connected to the input nodes.

    :param nodes1: the first child node
    :type nodes1: Union[ProdNodes,InputNodes]

    :param args: the remaining child nodes
    :type args: Union[ProdNodes,InputNodes]

    :param num_node_blocks: number of node blocks each of size `block_size`. I.e., the total number of nodes is `num_node_blocks` * `block_size`
    :type num_node_blocks: int

    :param num_nodes: an alternative of `num_node_blocks`; if set, it should be divicible by `block_size`
    :type num_nodes: int

    :param edge_ids: a matrix of size [2, # edges] - every size-2 column vector [i,j] defines a set of edges that fully connect the ith sum node block and the jth child node block
    :type edge_ids: Optional[Tensor]

    :param block_size: block size of the nodes; it does not change the semantics of the nodes, but will affect the speed of the compiled PC
    :type block_size: int

    :returns: an `SumNodes` object (a subclass of `CircuitNodes`)
    """

    assert block_size == 0 or block_size & (block_size - 1) == 0, "`block_size` must be a power of 2."

    if num_nodes > 0:
        assert num_node_blocks == 0, "Only one of `num_nodes` and `num_node_blocks` can be set at the same time."
        if block_size == 0:
            block_size = CircuitNodes.DEFAULT_BLOCK_SIZE

        assert num_nodes % block_size == 0

        num_node_blocks = num_nodes // block_size

    assert isinstance(nodes1, ProdNodes) or isinstance(nodes1, InputNodes), f"Children of sum nodes must be input or product nodes, but found input of type {type(nodes1)}." 

    if edge_ids is not None and num_node_blocks == 0:
        num_node_blocks = edge_ids[0,:].max().item() + 1

    assert num_node_blocks > 0, "Number of node blocks should be greater than 0."

    chs = [nodes1]
    scope = deepcopy(nodes1.scope)
    for nodes in args:
        assert isinstance(nodes, ProdNodes) or isinstance(nodes, InputNodes), f"Children of sum nodes must be input or product nodes, but found input of type {type(nodes)}."
        if not RegionGraph.ALLOW_NONSMOOTH:
            assert nodes.scope == scope, "Children of a `SumNodes` should have the same scope."
        chs.append(nodes)

    return SumNodes(num_node_blocks, chs, edge_ids, block_size = block_size, **kwargs)


class set_block_size(_DecoratorContextManager):
    """
    Context-manager that sets the block size of PC nodes constructed by `inputs`, `multiply`, and `summate`.

    :param block_size: the target block size
    :type block_size: int

    Example::
        >>> with pyjuice.set_block_size(16):
        ...     nis = pyjuice.inputs(var = 0, num_node_blocks = 4, dist = Categorical(num_cats = 20))
        >>> print(nis.num_nodes) # Should be 16 * 4
        64
    """

    def __init__(self, block_size: int = 1):

        assert block_size & (block_size - 1) == 0, "`block_size` must be a power of 2."

        self.block_size = block_size
        
        self.original_block_size = None

    def __enter__(self) -> None:
        self.original_block_size = CircuitNodes.DEFAULT_BLOCK_SIZE
        CircuitNodes.DEFAULT_BLOCK_SIZE = self.block_size

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        CircuitNodes.DEFAULT_BLOCK_SIZE = self.original_block_size


class structural_properties(_DecoratorContextManager):
    """
    Context-manager that controls the assertions of circuit structural properties, including smoothness and decomposability.

    :param allow_nonsmooth: whether to allow non-smooth circuits
    :type allow_nonsmooth: bool

    :param allow_nondecomposable: whether to allow non-decomposable circuits
    :type allow_nondecomposable: bool

    Example::
        >>> with pyjuice.structural_properties(allow_nonsmooth = True):
        ...     nis = pyjuice.inputs(var = 0, num_node_blocks = 4, dist = Categorical(num_cats = 20))
        ...     ....
    """

    def __init__(self, allow_nonsmooth = False, allow_nondecomposable = False):

        self.allow_nonsmooth = allow_nonsmooth
        self.allow_nondecomposable = allow_nondecomposable

    def __enter__(self) -> None:
        RegionGraph.ALLOW_NONSMOOTH = self.allow_nonsmooth
        RegionGraph.ALLOW_NONDECOMPOSABLE = self.allow_nondecomposable

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        RegionGraph.ALLOW_NONSMOOTH = False
        RegionGraph.ALLOW_NONDECOMPOSABLE = False



from __future__ import annotations

import torch
import torch.nn as nn
import triton
import triton.language as tl

from pyjuice.layer import SumLayer, ProdLayer, InputLayer
from pyjuice.utils.kernel_launcher import triton_jit

from .eval_partition import eval_partition_fn, prod_layer_partition_fn


@triton_jit
def prod_layer_td_backward_kernel(node_flows, element_flows, u_cids, parids, c_num_nblocks: tl.constexpr,
                                  c_num_nodes: tl.constexpr, n_num_nblocks: tl.constexpr, block_size: tl.constexpr,
                                  TILE_SIZE_N: tl.constexpr, TILE_SIZE_C: tl.constexpr):
    pid_c = tl.program_id(0)

    c_ids = tl.arange(0, TILE_SIZE_C) + pid_c * TILE_SIZE_C
    c_mask = (c_ids < c_num_nodes)
    group_ids = c_ids // block_size
    offs_node = c_ids % block_size

    n_ids = tl.arange(0, TILE_SIZE_N)
    n_mask = (n_ids < n_num_nblocks)

    cn_mask = c_mask[:,None] & n_mask[None,:]

    nids_base = tl.load(parids + group_ids[:,None] * n_num_nblocks + n_ids[None,:], mask = cn_mask)
    nids = nids_base + offs_node[:,None]
    nflows = tl.load(element_flows + nids, mask = cn_mask, other = 0.0)

    cflows = tl.sum(nflows, axis = 1)

    cids_base = tl.load(u_cids + group_ids, mask = c_mask)
    cids = cids_base + offs_node
    tl.atomic_add(node_flows + cids, cflows, mask = c_mask)


def prod_layer_td_backward(layer, node_flows, element_flows):
    block_size = layer.block_size
    for partition_id in range(layer.num_bk_partitions):
        u_cids = layer.partitioned_u_cids[partition_id]
        parids = layer.partitioned_parids[partition_id]

        c_num_nblocks = u_cids.size(0)
        c_num_nodes = c_num_nblocks * block_size
        n_num_nblocks = parids.size(1)

        if n_num_nblocks <= 2048:
            TILE_SIZE_N = triton.next_power_of_2(n_num_nblocks)
            TILE_SIZE_C = min(2048 // TILE_SIZE_N, triton.next_power_of_2(c_num_nodes))

            grid = (triton.cdiv(c_num_nodes, TILE_SIZE_C),)

            prod_layer_td_backward_kernel[grid](
                node_flows,
                element_flows,
                u_cids,
                parids,
                c_num_nblocks = c_num_nblocks,
                c_num_nodes = c_num_nodes,
                n_num_nblocks = n_num_nblocks,
                block_size = block_size,
                TILE_SIZE_N = TILE_SIZE_N,
                TILE_SIZE_C = TILE_SIZE_C
            )

        else:
            raise NotImplementedError("Need to implement a new kernel for this case.")


@triton_jit
def sum_layer_td_backward_block_kernel(node_flows, element_flows, node_mars, element_mars, mparams, chids, parids, parpids, 
                                       n_num_nodes: tl.constexpr, c_num_nodes: tl.constexpr, TILE_SIZE_N: tl.constexpr, 
                                       TILE_SIZE_C: tl.constexpr, n_block_size: tl.constexpr, c_block_size: tl.constexpr, 
                                       NUM_N_BLKS: tl.constexpr, pc_is_normalized: tl.constexpr):
    pid_n = tl.program_id(0)
    pid_c = tl.program_id(1)

    cgroup_id = (pid_c * TILE_SIZE_C) // c_block_size
    offs_cnode = tl.arange(0, TILE_SIZE_C) + ((pid_c * TILE_SIZE_C) % c_block_size)

    cid_base = tl.load(chids + cgroup_id)
    cids = cid_base + offs_cnode # [TILE_SIZE_C]

    ngroup_id = (pid_n * TILE_SIZE_N) // n_block_size
    offs_nnode = tl.arange(0, TILE_SIZE_N) + ((pid_n * TILE_SIZE_N) % n_block_size)

    nid_base = tl.load(parids + cgroup_id * NUM_N_BLKS + ngroup_id)
    nids = nid_base + offs_nnode # [TILE_SIZE_N]
    nflows = tl.load(node_flows + nids)

    pid_base = tl.load(parpids + cgroup_id * NUM_N_BLKS + ngroup_id)
    pids = pid_base + offs_cnode[:,None] * n_block_size + offs_nnode[None,:] # [TILE_SIZE_C, TILE_SIZE_N]
    epars = tl.load(mparams + pids)

    if not pc_is_normalized:
        cmars = tl.load(element_mars + cids) # [TILE_SIZE_C]
        nmars = tl.load(node_mars + nids) # [TILE_SIZE_N]

        epars = tl.exp(tl.log(epars) + cmars[:,None] - nmars[None,:])

    cflows = tl.sum(epars * nflows[None,:], axis = 1)
    tl.atomic_add(element_flows + cids, cflows)


@triton_jit
def sum_layer_td_backward_node_kernel(node_flows, element_flows, node_mars, element_mars, mparams, chids, parids, parpids, 
                                      n_num_nodes: tl.constexpr, c_num_nodes: tl.constexpr, TILE_SIZE_N: tl.constexpr, 
                                      TILE_SIZE_C: tl.constexpr, n_block_size: tl.constexpr, c_block_size: tl.constexpr, 
                                      NUM_N_BLKS: tl.constexpr, pc_is_normalized: tl.constexpr):
    pid_n = tl.program_id(0)
    pid_c = tl.program_id(1)

    c_ids = tl.arange(0, TILE_SIZE_C) + pid_c * TILE_SIZE_C
    c_mask = (c_ids < c_num_nodes)
    cgroup_ids = c_ids // c_block_size
    offs_cnode = c_ids % c_block_size

    cids_base = tl.load(chids + cgroup_ids)
    cids = cids_base + offs_cnode # [TILE_SIZE_C]
    
    n_ids = tl.arange(0, TILE_SIZE_N) + pid_n * TILE_SIZE_N
    n_mask = (n_ids < n_num_nodes)
    ngroup_ids = n_ids // n_block_size
    offs_nnode = n_ids % n_block_size

    cn_mask = c_mask[:,None] & n_mask[None,:]

    nids_base = tl.load(parids + cgroup_ids[:,None] * NUM_N_BLKS + ngroup_ids[None,:], mask = cn_mask)
    nids = nids_base + offs_nnode[None,:] # [TILE_SIZE_C, TILE_SIZE_N]
    nflows = tl.load(node_flows + nids, mask = cn_mask, other = 0.0)

    pids_base = tl.load(parpids + cgroup_ids[:,None] * NUM_N_BLKS + ngroup_ids[None,:], mask = cn_mask)
    pids = pids_base + offs_cnode[:,None] * n_block_size + offs_nnode[None,:] # [TILE_SIZE_C, TILE_SIZE_N]
    epars = tl.load(mparams + pids, mask = cn_mask, other = 0.0)

    if not pc_is_normalized:
        cmars = tl.load(element_mars + cids) # [TILE_SIZE_C]
        nmars = tl.load(node_mars + nids, mask = cn_mask, other = 0.0) # [TILE_SIZE_C, TILE_SIZE_N]

        epars = tl.exp(tl.log(epars) + cmars[:,None] - nmars)

    cflows = tl.sum(epars * nflows, axis = 1)
    tl.atomic_add(element_flows + cids, cflows)


def sum_layer_td_backward(layer, node_flows, element_flows, node_mars, element_mars, params, pc_is_normalized = True):
    block_size = layer.block_size
    for partition_id in range(layer.num_bk_partitions):
        chids = layer.partitioned_chids[partition_id]
        parids = layer.partitioned_parids[partition_id]
        parpids = layer.partitioned_parpids[partition_id]
        cs_block_size = layer.cs_block_sizes[partition_id]

        n_num_nodes = parids.size(1) * block_size # max number of parents per c node
        c_num_nodes = chids.size(0) * cs_block_size
        
        if block_size >= 8 and cs_block_size >= 8:
            TILE_SIZE_C = min(64, cs_block_size)
            TILE_SIZE_N = min(64, block_size)

            grid = (triton.cdiv(n_num_nodes, TILE_SIZE_N), triton.cdiv(c_num_nodes, TILE_SIZE_C))

            sum_layer_td_backward_block_kernel[grid](
                node_flows,
                element_flows,
                node_mars,
                element_mars,
                params,
                chids,
                parids,
                parpids,
                n_num_nodes = n_num_nodes,
                c_num_nodes = c_num_nodes,
                TILE_SIZE_N = TILE_SIZE_N,
                TILE_SIZE_C = TILE_SIZE_C,
                n_block_size = block_size,
                c_block_size = cs_block_size,
                NUM_N_BLKS = parids.size(1),
                pc_is_normalized = pc_is_normalized,
            )

        else:
            TILE_SIZE_C = min(32, triton.next_power_of_2(c_num_nodes))
            TILE_SIZE_N = min(32, triton.next_power_of_2(n_num_nodes))

            grid = (triton.cdiv(n_num_nodes, TILE_SIZE_N), triton.cdiv(c_num_nodes, TILE_SIZE_C))

            sum_layer_td_backward_node_kernel[grid](
                node_flows,
                element_flows,
                node_mars,
                element_mars,
                params,
                chids,
                parids,
                parpids,
                n_num_nodes = n_num_nodes,
                c_num_nodes = c_num_nodes,
                TILE_SIZE_N = TILE_SIZE_N,
                TILE_SIZE_C = TILE_SIZE_C,
                n_block_size = block_size,
                c_block_size = cs_block_size,
                NUM_N_BLKS = parids.size(1),
                pc_is_normalized = pc_is_normalized,
            )


@triton_jit
def sum_layer_td_pflow_kernel(node_flows, node_mars, element_mars, mparams, param_flows, nids, cids, pids, pfids, hyperparameters,
                              n_num_nodes: tl.constexpr, c_num_nodes: tl.constexpr, TILE_SIZE_N: tl.constexpr, TILE_SIZE_C: tl.constexpr,
                              n_block_size: tl.constexpr, pc_is_normalized: tl.constexpr):
    pid_c = tl.program_id(0)
    pid_n = tl.program_id(1)

    scale = tl.load(hyperparameters)

    n_ids = tl.arange(0, TILE_SIZE_N) + pid_n * TILE_SIZE_N
    n_mask = (n_ids < n_num_nodes)
    ngroup_ids = n_ids // n_block_size
    offs_nnode = n_ids % n_block_size

    nids_base = tl.load(nids + ngroup_ids, mask = n_mask)
    snids = nids_base + offs_nnode # [TILE_SIZE_N]
    nflows = tl.load(node_flows + snids, mask = n_mask, other = 0.0)

    offs_cnode = tl.arange(0, TILE_SIZE_C) + pid_c * TILE_SIZE_C
    c_mask = (offs_cnode < c_num_nodes)

    nc_mask = n_mask[:,None] & c_mask[None,:]

    pids_base = tl.load(pids + ngroup_ids[:,None] * c_num_nodes + offs_cnode[None,:], mask = nc_mask)
    epids = pids_base + offs_nnode[:,None] # [TILE_SIZE_N, TILE_SIZE_C]
    epars = tl.load(mparams + epids, mask = nc_mask, other = 0.0)

    if not pc_is_normalized:
        nmars = tl.load(node_mars + snids, mask = n_mask, other = 0.0)

        cids_base = tl.load(cids + ngroup_ids[:,None] * c_num_nodes + offs_cnode[None,:], mask = c_mask[None,:])
        cmars = tl.load(element_mars + cids_base, mask = c_mask[None,:], other = 0.0) # [TILE_SIZE_N, TILE_SIZE_C]

        epars = tl.exp(tl.log(epars) + cmars - nmars[:,None])

    eflows = epars * nflows[:,None] * scale

    pfids_base = tl.load(pfids + ngroup_ids[:,None] * c_num_nodes + offs_cnode[None,:], mask = nc_mask)
    epfids = pfids_base + offs_nnode[:,None] # [TILE_SIZE_N, TILE_SIZE_C]
    tl.atomic_add(param_flows + epfids, eflows, mask = nc_mask)


def sum_layer_td_pflow(layer, node_flows, node_mars, element_mars, params, param_flows, hyperparameters, pc_is_normalized = True):
    block_size = layer.block_size
    for partition_id in range(layer.num_fw_partitions):
        nids = layer.partitioned_nids[partition_id]
        cids = layer.partitioned_cids[partition_id]
        pids = layer.partitioned_pids[partition_id]
        pfids = layer.partitioned_pfids[partition_id]

        n_num_nodes = nids.size(0) * block_size
        c_num_nodes = cids.size(1)

        TILE_SIZE_N = min(32, triton.next_power_of_2(n_num_nodes))
        TILE_SIZE_C = min(32, triton.next_power_of_2(c_num_nodes))

        grid = (triton.cdiv(c_num_nodes, TILE_SIZE_C), triton.cdiv(n_num_nodes, TILE_SIZE_N))

        sum_layer_td_pflow_kernel[grid](
            node_flows,
            node_mars,
            element_mars,
            params,
            param_flows,
            nids,
            cids,
            pids,
            pfids,
            hyperparameters,
            n_num_nodes = n_num_nodes,
            c_num_nodes = c_num_nodes,
            TILE_SIZE_N = TILE_SIZE_N,
            TILE_SIZE_C = TILE_SIZE_C,
            n_block_size = block_size,
            pc_is_normalized = pc_is_normalized
        )


def eval_top_down_probs(pc, update_pflow: bool = True, scale: float = 1.0, pc_is_normalized: bool = True, use_cudagraph: bool = True):
    """
    Computes the top-down probabilities of every node.
    We use the first # nodes entries in `pc.node_flows` and the first # element entries in `pc.element_flows`.
    """
    assert hasattr(pc, "node_flows") and hasattr(pc, "element_flows")
    assert pc.node_flows.numel() >= pc.num_nodes
    assert pc.element_flows.numel() >= pc.num_elements

    if not hasattr(pc, "_single_node_flows"):
        pc._single_node_flows = torch.zeros([pc.num_nodes], dtype = torch.float32, device = pc.node_flows.device)

    if not hasattr(pc, "_single_element_flows"):
        pc._single_element_flows = torch.zeros([pc.num_elements], dtype = torch.float32, device = pc.element_flows.device)

    if not hasattr(pc, "_tdp_hyperparams"):
        pc._tdp_hyperparams = torch.zeros([1], dtype = torch.float32, device = pc.node_flows.device)

    node_flows = pc._single_node_flows
    element_flows = pc._single_element_flows
    hyperparameters = pc._tdp_hyperparams

    node_flows[:] = 0.0
    element_flows[:] = 0.0

    node_flows[pc._root_node_range[0]:pc._root_node_range[1]] = 1.0

    # Load hyperparameter(s)
    hyperparameters[0] = scale

    if not pc_is_normalized:
        # Run a forward pass to evaluate the partition function of every node
        eval_partition_fn(pc)
        node_mars = pc.node_mars.view(-1)
        element_mars = pc.element_mars.view(-1)
    else:
        node_mars, element_mars = None, None

    def run_tdp_backward():
        # Backward pass over the inner layers
        for layer_id in range(len(pc.inner_layer_groups) - 1, -1, -1):
            layer_group = pc.inner_layer_groups[layer_id]

            if not pc_is_normalized and layer_group.is_sum():
                prod_layer_group = pc.inner_layer_groups[layer_id]
                for layer in prod_layer_group:
                    prod_layer_partition_fn(layer, node_mars, element_mars, pc.params)

            for layer in layer_group:
                if layer.is_prod():
                    prod_layer_td_backward(layer, node_flows, element_flows)

                elif layer.is_sum():
                    element_flows[:] = 0.0
                    sum_layer_td_backward(layer, node_flows, element_flows, node_mars, element_mars, pc.params, 
                                          pc_is_normalized = pc_is_normalized)

                    if update_pflow:
                        sum_layer_td_pflow(layer, node_flows, node_mars, element_mars, pc.params, pc.param_flows, hyperparameters,
                                           pc_is_normalized = pc_is_normalized)

    if not hasattr(pc, "_tdp_cudagraph"):
        pc._tdp_cudagraph = dict()

    key = (update_pflow, pc_is_normalized, (None if node_mars is None else id(node_mars)), 
           (None if element_mars is None else id(element_mars)), id(node_flows), id(pc.params), 
           id(pc.param_flows), id(hyperparameters))
    if use_cudagraph and key in pc._tdp_cudagraph:
        g = pc._tdp_cudagraph[key]
        g.replay()
    elif not use_cudagraph:
        run_tdp_backward()
    else:
        # Backup param flows
        backup_node_flows = node_flows.detach().cpu().clone()
        backup_element_flows = element_flows.detach().cpu().clone()
        if update_pflow:
            backup_param_flows = pc.param_flows.detach().cpu().clone()
            backup_input_param_flows = []
            for layer in pc.input_layer_group:
                backup_input_param_flows.append(layer.param_flows.detach().cpu().clone())

        # Record CUDAGraph
        s = torch.cuda.Stream()
        s.wait_stream(torch.cuda.current_stream())
        with torch.cuda.stream(s):
            for _ in range(3):
                run_tdp_backward()
        torch.cuda.current_stream().wait_stream(s)

        g = torch.cuda.CUDAGraph()
        with torch.cuda.graph(g):
            run_tdp_backward()

        pc._tdp_cudagraph[key] = g

        # Restore param_flows (no extra GPU allocation)
        node_flows.copy_(backup_node_flows, non_blocking = True)
        element_flows.copy_(backup_element_flows, non_blocking = True)
        if update_pflow:
            pc.param_flows.copy_(backup_param_flows, non_blocking = True)
            for layer, backup_pfs in zip(pc.input_layer_group, backup_input_param_flows):
                layer.param_flows.copy_(backup_pfs, non_blocking = True)
        torch.cuda.synchronize()
        
        g.replay()

    # Backward pass over the input layers
    if update_pflow:
        for layer in pc.input_layer_group:
            layer.add_missing_flows(node_flows, scale = scale)

    return None



from __future__ import annotations

import torch
import triton
import triton.language as tl
from typing import Union, Callable
from functools import partial

from pyjuice.nodes import CircuitNodes
from pyjuice.model import TensorCircuit
from pyjuice.nodes.methods import get_subsumed_scopes
from pyjuice.utils import BitSet
from pyjuice.utils.kernel_launcher import FastJITFunction
from .base import query


## Categorical layer ##

@triton.jit
def _soft_evi_categorical_fw_kernel(data_ptr, node_mars_ptr, params_ptr, vids_ptr, psids_ptr, node_nchs_ptr, local_ids,
                                    sid: tl.constexpr, num_nodes: tl.constexpr, num_cats: tl.constexpr, 
                                    batch_size: tl.constexpr, partial: tl.constexpr, BLOCK_SIZE: tl.constexpr):
    pid = tl.program_id(axis = 0)
    block_start = pid * BLOCK_SIZE

    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < num_nodes * batch_size

    # Get node ID and category ID
    ns_offsets = offsets // batch_size
    batch_offsets = offsets % batch_size

    # Get number of children (categories)
    node_nch = tl.load(node_nchs_ptr + ns_offsets, mask = mask, other = 0)

    # Get variable ID
    vid = tl.load(vids_ptr + ns_offsets, mask = mask, other = 0)

    # Get param start ID
    psid = tl.load(psids_ptr + ns_offsets, mask = mask, other = 0)

    # Compute soft evidence per category
    node_vals = tl.zeros((BLOCK_SIZE,), tl.float32)
    for cat_id in range(num_cats):

        cmask = mask & (cat_id < node_nch)

        # Get data (soft evidence)
        data_offsets = vid * (num_cats * batch_size) + cat_id * batch_size + batch_offsets
        d_soft_evi = tl.load(data_ptr + data_offsets, mask = cmask, other = 0)

        # Get param
        param = tl.load(params_ptr + psid + cat_id, mask = cmask, other = 0)

        # Compute current likelihood and accumulate
        node_vals += d_soft_evi * param

    # Write back
    if not partial:
        tl.store(node_mars_ptr + offsets + (sid * batch_size), tl.log(node_vals), mask = mask)
    else:
        global_nid = tl.load(local_ids + ns_offsets, mask = mask, other = 0) + sid
        tl.store(node_mars_ptr + global_nid * batch_size + batch_offsets, tl.log(node_vals), mask = mask)


def _categorical_forward(layer, inputs: torch.Tensor, node_mars: torch.Tensor,
                         missing_mask: Optional[torch.Tensor] = None, **kwargs):

    batch_size, num_vars = inputs.size(0), inputs.size(1)

    if inputs.dim() == 2:
        # Hard evidence
        assert inputs.dtype == torch.long

        inputs = inputs.permute(1, 0).contiguous()

        layer.forward(data = inputs, node_mars = node_mars, missing_mask = missing_mask)

    elif inputs.dim() == 3:
        # Soft evidence
        assert inputs.dtype == torch.float32 and inputs.min() >= 0.0 and inputs.max() <= 1.0

        if missing_mask is not None:
            if missing_mask.dim() == 1:
                inputs[:,missing_mask,:] = 1.0
            else:
                assert missing_mask.dim() == 2
                inputs = inputs.flatten(0, 1)
                inputs[missing_mask.flatten(),:] = 1.0
                inputs = inputs.reshape(batch_size, num_vars, -1)

        inputs = inputs.permute(1, 2, 0) # [num_vars, num_cats, B]
        num_cats = inputs.size(1)

        sid, eid = layer._output_ind_range[0], layer._output_ind_range[1]
        num_nodes = eid - sid

        node_nchs = layer.metadata[layer.s_mids]

        grid = lambda meta: (triton.cdiv(num_nodes * batch_size, meta['BLOCK_SIZE']),)

        _soft_evi_categorical_fw_kernel[grid](
            inputs.reshape(-1).contiguous(), node_mars, layer.params, layer.vids.reshape(-1), layer.s_pids, node_nchs,
            None, sid, num_nodes, num_cats, batch_size, partial = False, BLOCK_SIZE = 512
        )

        node_mars[sid:eid,:] = node_mars[sid:eid,:].clip(max = 0.0)

    else:
        raise NotImplementedError("Unknown method to compute the forward pass for `Categorical`.")

    return None


@triton.jit
def _categorical_backward_kernel(cat_probs_ptr, node_flows_ptr, local_ids_ptr, rev_vars_mapping_ptr, vids_ptr, psids_ptr, 
                                 node_nchs_ptr, params_ptr, sid, eid, num_target_nodes, batch_size: tl.constexpr, 
                                 num_cats: tl.constexpr, partial_eval: tl.constexpr, BLOCK_SIZE: tl.constexpr):
    pid = tl.program_id(axis = 0)
    block_start = pid * BLOCK_SIZE

    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = (offsets < num_target_nodes * batch_size)

    # Get node offsets and batch offsets
    local_offsets = (offsets // batch_size)
    if partial_eval == 1: 
        local_node_offsets = tl.load(local_ids_ptr + local_offsets, mask = mask, other = 0)
    else:
        local_node_offsets = local_offsets
    batch_offsets = (offsets % batch_size)

    global_node_offsets = local_node_offsets + sid

    # Get variable ID
    origin_vid = tl.load(vids_ptr + local_node_offsets, mask = mask, other = 0)
    vid = tl.load(rev_vars_mapping_ptr + origin_vid, mask = mask, other = 0)

    # Get number of children per node
    node_nch = tl.load(node_nchs_ptr + local_node_offsets, mask = mask, other = 0)

    # Get param start ID
    psid = tl.load(psids_ptr + local_node_offsets, mask = mask, other = 0)

    # Get flow
    nflow_offsets = global_node_offsets * batch_size + batch_offsets
    nflow = tl.load(node_flows_ptr + nflow_offsets, mask = mask, other = 0)

    # Compute edge flows and accumulate
    for cat_id in range(num_cats):
        cmask = mask & (cat_id < node_nch)

        param = tl.load(params_ptr + psid + cat_id, mask = cmask, other = 0)
        eflow = nflow * param

        p_offsets = vid * num_cats * batch_size + cat_id * batch_size + batch_offsets
        tl.atomic_add(cat_probs_ptr + p_offsets, eflow, mask = cmask)


def _categorical_backward(layer, inputs: torch.Tensor, node_flows: torch.Tensor, node_mars: torch.Tensor,
                          params: Optional[torch.Tensor] = None, **kwargs):

    if params is None:
        params = layer.params

    sid, eid = layer._output_ind_range[0], layer._output_ind_range[1]

    num_nodes = eid - sid
    num_vars = layer.vids.max().item() + 1
    num_cats = int(layer.metadata[layer.s_mids].max().item())
    batch_size = node_flows.size(1)

    if "target_vars" in kwargs and kwargs["target_vars"] is not None:
        target_vars = kwargs["target_vars"]

        rev_vars_mapping = torch.zeros([num_vars], dtype = torch.long)
        for i, var in enumerate(target_vars):
            rev_vars_mapping[var] = i
        rev_vars_mapping = rev_vars_mapping.to(node_flows.device)
    else:
        target_vars = [var for var in range(num_vars)]

        rev_vars_mapping = torch.arange(0, num_vars, device = node_flows.device)

    num_target_vars = len(target_vars)

    cat_probs = torch.zeros([num_target_vars * num_cats * batch_size], dtype = torch.float32, device = node_flows.device)

    if len(target_vars) < num_vars:
        local_ids = layer.enable_partial_evaluation(bk_scopes = target_vars, return_ids = True).to(node_flows.device)
        num_target_nodes = local_ids.size(0)
        partial_eval = 1
    else:
        local_ids = None
        num_target_nodes = eid - sid
        partial_eval = 0

    node_nchs = layer.metadata[layer.s_mids]

    grid = lambda meta: (triton.cdiv(num_target_nodes * batch_size, meta['BLOCK_SIZE']),)

    _categorical_backward_kernel[grid](
        cat_probs, node_flows, local_ids, rev_vars_mapping, layer.vids, layer.s_pids, node_nchs, layer.params,
        sid, eid, num_target_nodes, batch_size, num_cats, partial_eval = partial_eval, BLOCK_SIZE = 512
    )

    cat_probs = cat_probs.reshape(num_target_vars, num_cats, batch_size)

    cat_probs /= (cat_probs.sum(dim = 1, keepdim = True) + 1e-12)
    cat_probs = cat_probs.permute(2, 0, 1)

    return cat_probs


def _conditional_fw_input_fn(layer, inputs, node_mars, **kwargs):
    if layer.dist_signature == "Categorical":
        _categorical_forward(layer, inputs, node_mars, **kwargs)

    else:
        raise TypeError(f"Unknown/unsupported layer type {type(layer)} for the forward pass. Please implement and provide your own `fw_input_fn`.")


def _conditional_bk_input_fn(layer, inputs, node_flows, node_mars, outputs = None, **kwargs):
    if layer.dist_signature == "Categorical":
        outputs.append(
            _categorical_backward(layer, inputs, node_flows, node_mars, layer.params, **kwargs)
        )

    else:
        raise TypeError(f"Unknown/unsupported layer type {type(layer)} for the backward pass. Please implement and provide your own `bk_input_fn`.")


## Main API ##


def conditional(pc: TensorCircuit, data: torch.Tensor, missing_mask: Optional[torch.Tensor] = None,
                target_vars: Optional[Sequence[int]] = None,
                fw_input_fn: Optional[Union[str,Callable]] = None, 
                bk_input_fn: Optional[Union[str,Callable]] = None, **kwargs):
    """
    Compute the conditional probability given hard or soft evidence, i.e., P(o|e).

    :param pc: the input PC
    :type pc: TensorCircuit

    :param data: data of size [B, num_vars] (hard evidence) or a custom shape paired with `fw_input_fn`
    :type data: torch.Tensor

    :param missing_mask: a boolean mask indicating marginalized variables; the size can be [num_vars] or [B, num_vars]
    :type missing_mask: torch.Tensor

    :param fw_input_fn: an optional custom function for the forward pass of input layers
    :type fw_input_fn: Optional[Union[str,Callable]]

    :param bk_input_fn: an optional custom function for the backward pass of input layers
    :type bk_input_fn: Optional[Union[str,Callable]]
    """

    outputs = []

    _wrapped_bk_input_fn = partial(_conditional_bk_input_fn, outputs = outputs)

    kwargs["target_vars"] = target_vars

    query(pc, inputs = data, run_backward = True, 
          fw_input_fn = _conditional_fw_input_fn if fw_input_fn is None else fw_input_fn, 
          bk_input_fn = _wrapped_bk_input_fn if bk_input_fn is None else bk_input_fn, 
          missing_mask = missing_mask, **kwargs)

    return outputs[0]



from __future__ import annotations

import torch
import triton
import triton.language as tl

from typing import Optional, Any

from .distributions import Distribution


class Categorical(Distribution):
    """
    A class representing Categorical distributions.

    :param num_cats: number of categories
    :type num_cats: int
    """
    def __init__(self, num_cats: int):
        super(Categorical, self).__init__()

        self.num_cats = num_cats

    def get_signature(self):
        """
        Get the signature of the current distribution.
        """
        return "Categorical"

    def get_metadata(self):
        """
        Get the metadata of the current distribution.
        """
        return [self.num_cats]

    def normalize_parameters(self, params: torch.Tensor):
        params = params.reshape(-1, self.num_cats)
        params /= params.sum(dim = 1, keepdim = True)

        return params.reshape(-1)

    def num_parameters(self):
        """
        The number of parameters per node.
        """
        return self.num_cats

    def num_param_flows(self):
        """
        The number of parameter flows per node.
        """
        return self.num_cats

    def init_parameters(self, num_nodes: int, perturbation: float = 2.0, params: Optional[Any] = None, **kwargs):
        """
        Initialize parameters for `num_nodes` nodes.
        Returned parameters should be flattened into a vector.
        """
        if params is not None:
            assert isinstance(params, torch.Tensor)
            assert params.numel() == num_nodes * self.num_parameters()
            return params

        params = torch.exp(torch.rand([num_nodes, self.num_cats]) * -perturbation)
        params /= params.sum(dim = 1, keepdim = True)

        return params.reshape(-1)

    def get_data_dtype(self):
        """
        Get the data dtype for the distribution.
        """
        return torch.long

    def get_em_fn(self):
        if self.num_cats <= 256:
            return self.small_ncats_em_fn
        else:
            self.em_block_size = 8
            return self.large_ncats_em_fn

    @staticmethod
    def fw_mar_fn(local_offsets, data, params_ptr, s_pids, metadata_ptr, s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE):
        # I am not sure why, but the following code will not work...
        # probs = tl.load(params_ptr + s_pids + data, mask = mask, other = 0)
        # Seems like a bug of triton.
        param_idx = s_pids + data
        probs = tl.load(params_ptr + param_idx, mask = mask, other = 0)
        log_probs = tl.log(probs)

        return log_probs

    @staticmethod
    def bk_flow_fn(local_offsets, ns_offsets, data, flows, node_mars_ptr, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, 
                   s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE):
        # I am not sure why, but the following code will not work...
        # tl.atomic_add(param_flows_ptr + s_pfids + data, flows, mask = mask)
        # Seems like a bug of triton.
        pf_offsets = s_pfids + data
        tl.atomic_add(param_flows_ptr + pf_offsets, flows, mask = mask)

    @staticmethod
    def bk_flow_mask_fn(local_offsets, ns_offsets, data, flows, node_mars_ptr, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, 
                        s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE, TILE_SIZE_K):
        # Get `num_cats` from `metadata`
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        num_cats = tl.load(metadata_ptr + s_mids, mask = mask, other = 0).to(tl.int64)

        max_num_cats = tl.max(num_cats, axis = 0)

        if TILE_SIZE_K > 1:
            num_iters = tlmath.ceil(max_num_cats / TILE_SIZE_K).to(tl.int64)

            cat_ids = tl.arange(0, TILE_SIZE_K)

            for i in range(num_iters):
                cat_mask = mask[:,None] & missing_mask[:,None] & (cat_ids[None,:] < num_cats[:,None])

                p_offsets = s_pids[:,None] + cat_ids[None,:]
                param = tl.load(params_ptr + p_offsets, mask = cat_mask, other = 0)

                pf_offsets = s_pfids[:,None] + cat_ids[None,:]
                tl.atomic_add(param_flows_ptr + pf_offsets, flows[:,None] * param, mask = cat_mask)

                cat_ids += TILE_SIZE_K
        else:
            for cat_id in range(max_num_cats):
                cat_mask = mask & missing_mask & (cat_id < num_cats)

                p_offsets = s_pids + cat_id
                param = tl.load(params_ptr + p_offsets, mask = cat_mask, other = 0)

                pf_offsets = s_pfids + cat_id
                tl.atomic_add(param_flows_ptr + pf_offsets, flows * param, mask = cat_mask)

    @staticmethod
    def sample_fn(samples_ptr, local_offsets, batch_offsets, vids, s_pids, params_ptr, metadata_ptr, s_mids_ptr, mask, batch_size, BLOCK_SIZE, seed):
        # Get `num_cats` from `metadata`
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        num_cats = tl.load(metadata_ptr + s_mids, mask = mask, other = 0).to(tl.int64)

        max_num_cats = tl.max(num_cats, axis = 0)

        rnd_val = tl.rand(seed, tl.arange(0, BLOCK_SIZE))
        sampled_id = tl.zeros([BLOCK_SIZE], dtype = tl.int64) - 1

        # Sample by computing cumulative probability
        cum_param = tl.zeros([BLOCK_SIZE], dtype = tl.float32)
        for cat_id in range(max_num_cats):
            cat_mask = mask & (cat_id < num_cats)

            param = tl.load(params_ptr + s_pids + cat_id, mask = cat_mask, other = 0)
            cum_param += param

            sampled_id = tl.where((cum_param >= rnd_val) & (sampled_id == -1), cat_id, sampled_id)

        sampled_id = tl.where((sampled_id == -1), 0, sampled_id)

        # Write back to `samples`
        sample_offsets = vids * batch_size + batch_offsets
        tl.store(samples_ptr + sample_offsets, sampled_id, mask = mask)

    @staticmethod
    def small_ncats_em_fn(local_offsets, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, s_mids_ptr, mask,
              step_size, pseudocount, BLOCK_SIZE):
        # Get `num_cats` from `metadata`
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        num_cats = tl.load(metadata_ptr + s_mids, mask = mask, other = 0).to(tl.int64)

        max_num_cats = tl.max(num_cats, axis = 0)

        # Compute cumulative flows
        cum_flow = tl.zeros([BLOCK_SIZE], dtype = tl.float32)
        for cat_id in range(max_num_cats):
            cat_mask = mask & (cat_id < num_cats)

            flow = tl.load(param_flows_ptr + s_pfids + cat_id, mask = cat_mask, other = 0)
            cum_flow += flow

        # Parameter update
        numerate_pseudocount = pseudocount / num_cats
        cum_flow += pseudocount
        for cat_id in range(max_num_cats):
            cat_mask = mask & (cat_id < num_cats)

            param = tl.load(params_ptr + s_pids + cat_id, mask = cat_mask, other = 0)
            flow = tl.load(param_flows_ptr + s_pfids + cat_id, mask = cat_mask, other = 0)

            new_param = (1.0 - step_size) * param + step_size * (flow + numerate_pseudocount) / cum_flow
            tl.store(params_ptr + s_pids + cat_id, new_param, mask = cat_mask)

    @staticmethod
    def large_ncats_em_fn(local_offsets, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, s_mids_ptr, mask,
              step_size, pseudocount, BLOCK_SIZE):
        # Get `num_cats` from `metadata`
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        num_cats = tl.load(metadata_ptr + s_mids, mask = mask, other = 0).to(tl.int64)

        max_num_cats = tl.max(num_cats, axis = 0)

        # Compute cumulative flows
        cum_flow = tl.zeros([BLOCK_SIZE], dtype = tl.float32)
        cat_ids = tl.arange(0, 128)
        for cat_sid in range(0, max_num_cats, 128):
            cat_mask = mask[:,None] & (cat_ids[None,:] < num_cats[:,None])

            flow = tl.load(param_flows_ptr + s_pfids[:,None] + cat_ids[None,:], mask = cat_mask, other = 0)
            cum_flow += tl.sum(flow, axis = 1)

            cat_ids += 128

        # Parameter update
        numerate_pseudocount = pseudocount / num_cats
        cum_flow += pseudocount
        cat_ids = tl.arange(0, 128)
        for cat_sid in range(0, max_num_cats, 128):
            cat_mask = mask[:,None] & (cat_ids[None,:] < num_cats[:,None])

            param = tl.load(params_ptr + s_pids[:,None] + cat_ids[None,:], mask = cat_mask, other = 0)
            flow = tl.load(param_flows_ptr + s_pfids[:,None] + cat_ids[None,:], mask = cat_mask, other = 0)

            new_param = (1.0 - step_size) * param + step_size * (flow + numerate_pseudocount[:,None]) / cum_flow[:,None]
            tl.store(params_ptr + s_pids[:,None] + cat_ids[None,:], new_param, mask = cat_mask)

            cat_ids += 128

    @staticmethod
    def partition_fn(local_offsets, params_ptr, s_pids, metadata_ptr, s_mids_ptr, mask, BLOCK_SIZE, TILE_SIZE_K):
        # Get `num_cats` from `metadata`
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        num_cats = tl.load(metadata_ptr + s_mids, mask = mask, other = 0).to(tl.int64)

        max_num_cats = tl.max(num_cats, axis = 0)

        num_iters = tlmath.ceil(max_num_cats / TILE_SIZE_K).to(tl.int64)

        cat_ids = tl.arange(0, TILE_SIZE_K)
        mars = tl.zeros([BLOCK_SIZE], dtype = tl.float32)

        for i in range(num_iters):
            cat_mask = mask[:,None] & (cat_ids[None,:] < num_cats[:,None])

            p_offsets = s_pids[:,None] + cat_ids[None,:]
            param = tl.load(params_ptr + p_offsets, mask = cat_mask, other = 0)

            mars += tl.sum(param, axis = 1)

            cat_ids += TILE_SIZE_K

        mars = tl.log(mars)
        return mars

    def _get_constructor(self):
        return Categorical, {"num_cats": self.num_cats}

    def _need_2nd_kernel_dim(self):
        return True



from __future__ import annotations

import torch
import triton
import triton.language as tl
import math
from typing import Tuple, Optional, Any

from .distributions import Distribution
from pyjuice.utils.kernel_launcher import triton_jit


def _condition_apply_soft_evi_kernel(layer, kwargs):
    return "external_soft_evi" in kwargs


def _prep_args_apply_soft_evi_kernel(layer, kwargs):
    target_kwargs = dict()

    assert "external_soft_evi" in kwargs
    external_soft_evi = kwargs["external_soft_evi"]
    assert external_soft_evi.dim() == 3

    target_kwargs["external_soft_evi_ptr"] = external_soft_evi

    target_kwargs["var_idmapping_ptr"] = layer.var_idmapping

    target_kwargs["ext_num_vars"] = external_soft_evi.size(1)

    target_kwargs["max_num_latents"] = external_soft_evi.size(2)

    return target_kwargs


def _condition_soft_evi_grad_kernel(layer, kwargs):
    return "external_soft_evi_grad" in kwargs


def _prep_args_soft_evi_grad_kernel(layer, kwargs):
    target_kwargs = dict()

    assert "external_soft_evi_grad" in kwargs
    external_soft_evi_grad = kwargs["external_soft_evi_grad"]
    assert external_soft_evi_grad.dim() == 3

    target_kwargs["external_soft_evi_grad_ptr"] = external_soft_evi_grad

    target_kwargs["var_idmapping_ptr"] = layer.var_idmapping

    target_kwargs["ext_num_vars"] = external_soft_evi_grad.size(1)

    target_kwargs["max_num_latents"] = external_soft_evi_grad.size(2)

    return target_kwargs


class External(Distribution):
    """
    A class representing user-define distributions (PyJuice only processes the incoming log-probabilities).

    :param val_range: range of the values represented by the distribution
    :type val_range: Tuple[float,float]

    :param num_cats: number of categories
    :type num_cats: int

    :param min_std: minimum standard deviation
    :type min_std: float
    """

    def __init__(self):
        super(External, self).__init__()

        self.post_fw_fns = [
            (self.apply_soft_evi_kernel, _condition_apply_soft_evi_kernel, _prep_args_apply_soft_evi_kernel)
        ]

        self.post_bp_fns = [
            (self.soft_evi_grad_kernel, _condition_soft_evi_grad_kernel, _prep_args_soft_evi_grad_kernel)
        ]

    def get_signature(self):
        """
        Get the signature of the current distribution.
        """
        return "External"

    def get_metadata(self):
        """
        Get the metadata of the current distribution.
        """
        return []

    def num_parameters(self):
        """
        The number of parameters per node.
        """
        return 0

    def num_param_flows(self):
        """
        The number of parameter flows per node.
        """
        return 0

    def init_parameters(self, num_nodes: int, perturbation: float = 2.0, params: Optional[Any] = None, **kwargs):
        """
        Initialize parameters for `num_nodes` nodes.
        Returned parameters should be flattened into a vector.
        """
        return torch.zeros([0], dtype = torch.float32)

    def get_data_dtype(self):
        """
        Get the data dtype for the distribution.
        """
        return torch.float32

    @staticmethod
    @triton_jit
    def apply_soft_evi_kernel(params_ptr, node_mars_ptr, data_ptr, vids_ptr, s_pids_ptr, metadata_ptr, s_mids_ptr, nids_ptr, 
                              fw_local_ids_ptr, partial_eval: tl.constexpr, layer_num_nodes: tl.constexpr, batch_size: tl.constexpr, 
                              num_vars_per_node: tl.constexpr, nv_block_size: tl.constexpr, node_offset: tl.constexpr, BLOCK_SIZE: tl.constexpr,
                              external_soft_evi_ptr, var_idmapping_ptr, ext_num_vars: tl.constexpr, max_num_latents: tl.constexpr):
        pid = tl.program_id(axis = 0)
        block_start = pid * BLOCK_SIZE

        offsets = block_start + tl.arange(0, BLOCK_SIZE)
        mask = offsets < layer_num_nodes * batch_size

        # Raw batch and (local) node id
        batch_offsets = (offsets % batch_size)
        local_offsets = (offsets // batch_size)

        if partial_eval > 0:
            local_offsets = tl.load(fw_local_ids_ptr + local_offsets, mask = mask, other = 0)

        # Get all variable ids
        vids = tl.load(vids_ptr + local_offsets, mask = mask, other = 0)
        lvids = tl.load(var_idmapping_ptr + vids, mask = mask, other = 0)

        # Get all latent offsets
        nids = tl.load(nids_ptr + local_offsets, mask = mask, other = 0)

        # Load the corresponding data
        soft_evi = tl.load(external_soft_evi_ptr + batch_offsets * (ext_num_vars * max_num_latents) + lvids * max_num_latents + nids, mask = mask, other = 0.0)

        node_offsets = local_offsets + node_offset
        tl.store(node_mars_ptr + node_offsets * batch_size + batch_offsets, soft_evi, mask = mask)

    @staticmethod
    @triton_jit
    def soft_evi_grad_kernel(params_ptr, param_flows_ptr, node_flows_ptr, node_mars_ptr, data_ptr, vids_ptr, s_pids_ptr, s_pfids_ptr,
                             metadata_ptr, s_mids_ptr, nids_ptr, bk_local_ids_ptr, partial_eval: tl.constexpr, logspace_flows: tl.constexpr, layer_num_nodes: tl.constexpr, 
                             batch_size: tl.constexpr, num_vars_per_node: tl.constexpr, num_vars: tl.constexpr, nv_block_size: tl.constexpr, node_offset: tl.constexpr, 
                             BLOCK_SIZE: tl.constexpr, TILE_SIZE_K: tl.constexpr, external_soft_evi_grad_ptr, var_idmapping_ptr, ext_num_vars: tl.constexpr, max_num_latents: tl.constexpr):
        pid = tl.program_id(axis = 0)
        block_start = pid * BLOCK_SIZE

        offsets = block_start + tl.arange(0, BLOCK_SIZE)
        mask = offsets < layer_num_nodes * batch_size

        # Raw batch and (local) node id
        batch_offsets = (offsets % batch_size)
        local_offsets = (offsets // batch_size)

        if partial_eval > 0:
            local_offsets = tl.load(bk_local_ids_ptr + local_offsets, mask = mask, other = 0)

        # Get all variable ids
        vids = tl.load(vids_ptr + local_offsets, mask = mask, other = 0)
        lvids = tl.load(var_idmapping_ptr + vids, mask = mask, other = 0)

        # Get all latent offsets
        nids = tl.load(nids_ptr + local_offsets, mask = mask, other = 0)

        # Load the flows
        ns_offsets = (local_offsets + node_offset) * batch_size + batch_offsets
        flows = tl.load(node_flows_ptr + ns_offsets, mask = mask, other = 0)

        # Store the corresponding data
        tl.store(external_soft_evi_grad_ptr + batch_offsets * (ext_num_vars * max_num_latents) + lvids * max_num_latents + nids, flows, mask = mask)

    @staticmethod
    def fw_mar_fn(local_offsets, data, params_ptr, s_pids, metadata_ptr, s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE):
        log_probs = tl.where(data == 0, 0.0, 0.0)

        return log_probs

    @staticmethod
    def bk_flow_fn(local_offsets, ns_offsets, data, flows, node_mars_ptr, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, 
                   s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE):
        pass

    @staticmethod
    def sample_fn(samples_ptr, local_offsets, batch_offsets, vids, s_pids, params_ptr, metadata_ptr, s_mids_ptr, mask, batch_size, BLOCK_SIZE, seed):
        pass

    @staticmethod
    def em_fn(local_offsets, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, s_mids_ptr, mask,
              step_size, pseudocount, BLOCK_SIZE):
        pass

    def _get_constructor(self):
        return External, {}


from .parflow_fusing import compile_cum_par_flows_fn, compute_cum_par_flows, cum_par_flows_to_device
from .par_update import compile_par_update_fn, em_par_update, par_update_to_device, sgd_par_update
from .normalize import normalize_parameters
from .eval_partition import eval_partition_fn
from .top_down_prob import eval_top_down_probs



from __future__ import annotations

import torch
from typing import Union, Callable

from pyjuice.nodes import CircuitNodes
from pyjuice.model import TensorCircuit


def query(pc: TensorCircuit, inputs: torch.Tensor, 
          run_backward: bool = True,
          fw_input_fn: Optional[Union[str,Callable]] = None, 
          bk_input_fn: Optional[Union[str,Callable]] = None, 
          fw_output_fn: Optional[Callable] = None, **kwargs):

    # Run forward pass
    lls = pc.forward(inputs, input_layer_fn = fw_input_fn, **kwargs)

    if fw_output_fn is not None:
        return fw_output_fn(pc)
    elif not run_backward:
        return lls

    # (Optionally) run backward pass
    assert bk_input_fn is not None
    pc.backward(inputs, input_layer_fn = bk_input_fn, compute_param_flows = False, **kwargs)

    return None


import torch
import triton
import triton.language as tl

from typing import Optional


@triton.jit
def _cum_params_kernel(params_ptr, cum_params_ptr, node_ids_ptr, num_param_blocks, block_size, batch_size, 
                       BLOCK_M: tl.constexpr, BLOCK_K: tl.constexpr, BLOCK_B: tl.constexpr):

    b_pid = tl.program_id(axis = 0)
    k_pid = tl.program_id(axis = 1)
    m_pid = tl.program_id(axis = 2)

    m_offsets = m_pid * BLOCK_M + tl.arange(0, BLOCK_M)
    m_mask = m_offsets < num_param_blocks

    k_offsets = k_pid * BLOCK_K + tl.arange(0, BLOCK_K)

    b_offsets = b_pid * BLOCK_B + tl.arange(0, BLOCK_B)
    b_mask = b_offsets < batch_size

    n_offsets = tl.load(node_ids_ptr + m_offsets, mask = m_mask, other = 0)
    reuse_offs = k_offsets[None,:,None] * batch_size + b_offsets[None,None,:]

    n_offsets = n_offsets[:,None,None] * (batch_size * block_size) + reuse_offs
    p_offsets = m_offsets[:,None,None] * (batch_size * block_size) + reuse_offs

    mask = m_mask[:,None,None] & b_mask[None,None,:]
    params = tl.load(params_ptr + p_offsets, mask = mask, other = 0)

    tl.atomic_add(cum_params_ptr + n_offsets, params, mask = mask)


@triton.jit
def _norm_params_kernel(params_ptr, cum_params_ptr, node_ids_ptr, node_nchs_ptr, num_param_blocks, block_size, 
                        batch_size, pseudocount, BLOCK_M: tl.constexpr, BLOCK_K: tl.constexpr, BLOCK_B: tl.constexpr):

    b_pid = tl.program_id(axis = 0)
    k_pid = tl.program_id(axis = 1)
    m_pid = tl.program_id(axis = 2)

    m_offsets = m_pid * BLOCK_M + tl.arange(0, BLOCK_M)
    m_mask = m_offsets < num_param_blocks

    k_offsets = k_pid * BLOCK_K + tl.arange(0, BLOCK_K)

    b_offsets = b_pid * BLOCK_B + tl.arange(0, BLOCK_B)
    b_mask = b_offsets < batch_size

    n_offsets = tl.load(node_ids_ptr + m_offsets, mask = m_mask, other = 0)
    reuse_offs = k_offsets[None,:,None] * batch_size + b_offsets[None,None,:]

    nb_offsets = n_offsets[:,None,None] * (batch_size * block_size) + reuse_offs
    p_offsets = m_offsets[:,None,None] * (batch_size * block_size) + reuse_offs

    mask = m_mask[:,None,None] & b_mask[None,None,:]
    params = tl.load(params_ptr + p_offsets, mask = mask, other = 0)
    cum_params = tl.load(cum_params_ptr + nb_offsets, mask = mask, other = 1)
    nchs = tl.load(node_nchs_ptr + n_offsets, mask = m_mask, other = 1)[:,None,None]
    
    normed_params = (params + pseudocount / nchs) / (cum_params + pseudocount)
    tl.store(params_ptr + p_offsets, normed_params, mask = mask)


def normalize_ns_parameters(params: torch.Tensor, node_ids: torch.Tensor, block_size: int, ch_block_size: int, 
                            node_nchs: Optional[torch.Tensor] = None, pseudocount: float = 0.0):

    assert 3 <= params.dim() <= 4 and params.size(1) == block_size and params.size(2) == ch_block_size

    num_param_blocks = params.size(0)
    num_node_blocks = torch.max(node_ids).detach().cpu().item() + 1

    if node_nchs is None:
        node_nchs = torch.bincount(node_ids) * ch_block_size

    if node_ids.is_cuda:
        assert params.is_cuda, "Input `params` should be on GPU."

        if params.dim() == 3:
            params = params.unsqueeze(3)

        batch_size = params.size(3)

        cum_params = torch.zeros([num_node_blocks, block_size, batch_size], dtype = torch.float32, device = params.device)

        blockified_params = params.sum(2).contiguous()

        BLOCK_B = min(batch_size, 128)
        BLOCK_K = min(1024 // BLOCK_B, triton.next_power_of_2(block_size))
        BLOCK_M = min(1024 // (BLOCK_B * BLOCK_K), triton.next_power_of_2(num_param_blocks))

        grid = lambda meta: (triton.cdiv(batch_size, BLOCK_B), triton.cdiv(block_size, BLOCK_K), triton.cdiv(num_param_blocks, BLOCK_M))

        _cum_params_kernel[grid](blockified_params, cum_params, node_ids, num_param_blocks, block_size, batch_size, BLOCK_M, BLOCK_K, BLOCK_B)
        _norm_params_kernel[grid](blockified_params, cum_params, node_ids, node_nchs, num_param_blocks, block_size, batch_size, pseudocount, BLOCK_M, BLOCK_K, BLOCK_B)

        params *= (blockified_params / (params.sum(2) + 1e-12)).unsqueeze(2)

    else:
        assert params.dim() == 3, "CPU version of `normalize_parameters` does not support `batch_size > 1` for now."

        with torch.no_grad():

            blockified_params = params.sum(dim = 2).contiguous()

            param_ids = torch.arange(0, num_param_blocks, dtype = torch.long, device = params.device)

            cum_matrix1 = torch.sparse_coo_tensor(
                torch.stack((node_ids, param_ids), dim = 0), 
                torch.ones([num_param_blocks], device = params.device), 
                (num_node_blocks, num_param_blocks)
            )
            node_buffer = torch.sparse.mm(cum_matrix1, blockified_params) + pseudocount

            node_buffer.reciprocal_()
            node_buffer = node_buffer.reshape(num_node_blocks * block_size, 1)

            param_ids = torch.arange(0, num_param_blocks * block_size, dtype = torch.long, device = params.device)
            flattened_node_ids = (node_ids.unsqueeze(1).repeat(1, block_size) * block_size + torch.arange(0, block_size, device = params.device)).reshape(-1)

            cum_matrix2 = torch.sparse_coo_tensor(
                torch.stack((param_ids, flattened_node_ids), dim = 0), 
                (blockified_params + pseudocount / node_nchs[node_ids].unsqueeze(1)).reshape(-1), 
                (num_param_blocks * block_size, num_node_blocks * block_size)
            )
            params_buffer = torch.sparse.mm(cum_matrix2, node_buffer).reshape(num_param_blocks, block_size)
            
            params *= (params_buffer / torch.where(blockified_params == 0.0, 1e-12, blockified_params)).unsqueeze(2)


from __future__ import annotations

import torch
import triton
import triton.language as tl
import math
from typing import Tuple, Optional, Any

from .distributions import Distribution


class DiscreteLogistic(Distribution):
    """
    A class representing Discrete Logistic distributions.

    :param val_range: range of the values represented by the distribution
    :type val_range: Tuple[float,float]

    :param num_cats: number of categories
    :type num_cats: int

    :param min_std: minimum standard deviation
    :type min_std: float
    """

    def __init__(self, val_range: Tuple[float,float], num_cats: int, min_std: float = 0.01):
        super(DiscreteLogistic, self).__init__()

        self.val_range = val_range
        self.num_cats = num_cats
        self.min_std = min_std

    def get_signature(self):
        """
        Get the signature of the current distribution.
        """
        return "DiscreteLogistic"

    def get_metadata(self):
        """
        Get the metadata of the current distribution.
        """
        return [self.val_range[0], self.val_range[1], self.num_cats, self.min_std]

    def num_parameters(self):
        """
        The number of parameters per node.
        """
        return 2

    def num_param_flows(self):
        """
        The number of parameter flows per node.
        """
        return 3

    def init_parameters(self, num_nodes: int, perturbation: float = 2.0, params: Optional[Any] = None, **kwargs):
        """
        Initialize parameters for `num_nodes` nodes.
        Returned parameters should be flattened into a vector.
        """
        if params is not None:
            assert isinstance(params, torch.Tensor)
            assert params.numel() == num_nodes * self.num_parameters()
            return params

        mu = kwargs["mu"] if "mu" in kwargs else (self.val_range[0] + self.val_range[1]) / 2.0
        std = kwargs["std"] if "std" in kwargs else (self.val_range[1] - self.val_range[0]) / 4.0 # 2 sigma range

        mus = torch.normal(mean = torch.ones([num_nodes]) * mu, std = std)
        ss = torch.exp(torch.rand([num_nodes]) * -0.04 * perturbation) * std * math.sqrt(3) / math.pi

        return torch.stack((mus, ss), dim = 1).reshape(-1).contiguous()

    def get_data_dtype(self):
        """
        Get the data dtype for the distribution.
        """
        return torch.float32

    @staticmethod
    def fw_mar_fn(local_offsets, data, params_ptr, s_pids, metadata_ptr, s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE):
        # Get `val_range` and `num_cats` from `metadata`
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        range_low = tl.load(metadata_ptr + s_mids, mask = mask, other = 0)
        range_high = tl.load(metadata_ptr + s_mids + 1, mask = mask, other = 0)
        num_cats = tl.load(metadata_ptr + s_mids + 2, mask = mask, other = 0).to(tl.int64)

        mu = tl.load(params_ptr + s_pids, mask = mask, other = 0)
        s = tl.load(params_ptr + s_pids + 1, mask = mask, other = 0)

        # 0: (-inf, range_low + interval]; 
        # 1: (range_low + interval, range_low + 2 * interval]; 
        # ...; 
        # num_cats - 1: [range_high - interval, inf)
        interval = (range_high - range_low) / num_cats
        vlow = data * interval + range_low
        vhigh = vlow + interval

        cdfhigh = tl.where(data == num_cats - 1, 1.0, 1.0 / (1.0 + tl.exp((mu - vhigh) / s)))
        cdflow = tl.where(data == 0, 0.0, 1.0 / (1.0 + tl.exp((mu - vlow) / s)))

        log_probs = tl.maximum(tl.log(cdfhigh - cdflow), -1000.0)

        return log_probs

    @staticmethod
    def bk_flow_fn(local_offsets, ns_offsets, data, flows, node_mars_ptr, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, 
                   s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE):
        # Get `val_range` and `num_cats` from `metadata`
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        range_low = tl.load(metadata_ptr + s_mids, mask = mask, other = 0)
        range_high = tl.load(metadata_ptr + s_mids + 1, mask = mask, other = 0)
        num_cats = tl.load(metadata_ptr + s_mids + 2, mask = mask, other = 0).to(tl.int64)

        interval = (range_high - range_low) / num_cats
        vmid = data * interval + range_low + 0.5 * interval # (vlow + vhigh) / 2

        stat1 = vmid * flows
        stat2 = vmid * vmid * flows
        stat3 = flows

        tl.atomic_add(param_flows_ptr + s_pfids, stat1, mask = mask)
        tl.atomic_add(param_flows_ptr + s_pfids + 1, stat2, mask = mask)
        tl.atomic_add(param_flows_ptr + s_pfids + 2, stat3, mask = mask)

    @staticmethod
    def bk_flow_mask_fn(local_offsets, ns_offsets, data, flows, node_mars_ptr, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, 
                        s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE, TILE_SIZE_K):
        mu = tl.load(params_ptr + s_pids, mask = mask, other = 0)
        s = tl.load(params_ptr + s_pids + 1, mask = mask, other = 0)

        stat1 = mu * flows
        stat2 = (s * s * 3.2898681337 + mu * mu) * flows # the constant is pi^2 / 3
        stat3 = flows

        tl.atomic_add(param_flows_ptr + s_pfids, stat1, mask = mask)
        tl.atomic_add(param_flows_ptr + s_pfids + 1, stat2, mask = mask)
        tl.atomic_add(param_flows_ptr + s_pfids + 2, stat3, mask = mask)

    @staticmethod
    def sample_fn(samples_ptr, local_offsets, batch_offsets, vids, s_pids, params_ptr, metadata_ptr, s_mids_ptr, mask, batch_size, BLOCK_SIZE, seed):
        # Get `val_range` and `num_cats` from `metadata`
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        range_low = tl.load(metadata_ptr + s_mids, mask = mask, other = 0)
        range_high = tl.load(metadata_ptr + s_mids + 1, mask = mask, other = 0)
        num_cats = tl.load(metadata_ptr + s_mids + 2, mask = mask, other = 0).to(tl.int64)

        rnd_val = tl.rand(seed, tl.arange(0, BLOCK_SIZE))

        mu = tl.load(params_ptr + s_pids, mask = mask, other = 0)
        s = tl.load(params_ptr + s_pids + 1, mask = mask, other = 0)
        
        # Get sample using the quantile function
        sample = mu + s * tl.log(rnd_val / (1.0 - rnd_val))

        # Discretize
        interval = (range_high - range_low) / num_cats
        bin_val = (sample - range_low) / interval
        sampled_id = tl.where(bin_val < 0, 0, tl.where(bin_val > num_cats - 1, num_cats - 1, tl.floor(bin_val).to(tl.int64)))

        # Write back to `samples`
        sample_offsets = vids * batch_size + batch_offsets
        tl.store(samples_ptr + sample_offsets, sampled_id, mask = mask)

    @staticmethod
    def em_fn(local_offsets, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, s_mids_ptr, mask,
              step_size, pseudocount, BLOCK_SIZE):
        # Get `min_std`
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        min_std = tl.load(metadata_ptr + s_mids + 3, mask = mask, other = 0)

        mu = tl.load(params_ptr + s_pids, mask = mask, other = 0)
        std = tl.load(params_ptr + s_pids + 1, mask = mask, other = 0) * 1.8137993642342178 # The constant is `math.pi / math.sqrt(3.0)`
        ori_theta1 = mu
        ori_theta2 = std * std + mu * mu

        stat1 = tl.load(param_flows_ptr + s_pfids, mask = mask)
        stat2 = tl.load(param_flows_ptr + s_pfids + 1, mask = mask)
        stat3 = tl.load(param_flows_ptr + s_pfids + 2, mask = mask)

        new_theta1 = stat1 / (stat3 + 1e-10)
        new_theta2 = stat2 / (stat3 + 1e-10)

        # Get the updated natural parameters
        updated_theta1 = (1.0 - step_size) * ori_theta1 + step_size * new_theta1
        updated_theta2 = (1.0 - step_size) * ori_theta2 + step_size * new_theta2

        # Reconstruct `mu` and `std` from the expectation parameters (moment matching)
        updated_mu = updated_theta1
        updated_std2 = updated_theta2 - updated_mu * updated_mu
        updated_std = tl.where(updated_std2 < min_std * min_std, min_std, tl.sqrt(updated_std2))
        updated_s = updated_std * 0.5513288954217921 # The constant is `math.sqrt(3.0) / math.pi`

        tl.store(params_ptr + s_pids, updated_mu, mask = mask)
        tl.store(params_ptr + s_pids + 1, updated_s, mask = mask)

    def _get_constructor(self):
        return DiscreteLogistic, {"val_range": self.val_range, "num_cats": self.num_cats, "min_std": self.min_std}



import torch
import triton
from typing import Callable, Tuple, Union


class FastJITFunction():
    def __init__(self, fn: Callable, device_check: bool = False):
        self.jit_fn = triton.JITFunction(fn)

        self.device_check = device_check

        try:
            self.constexpr_ids = [p.num for p in self.jit_fn.params if p.is_constexpr]
            self.constexpr_names = {p.name: p.num for p in self.jit_fn.params if p.is_constexpr}
            self.nonconstexpr_names = [p.name for p in self.jit_fn.params if not p.is_constexpr]
        except AttributeError:
            self.constexpr_ids = self.jit_fn.constexprs
            self.constexpr_names = {self.jit_fn.arg_names[i]: i for i in self.jit_fn.constexprs}
            self.nonconstexpr_names = [self.jit_fn.arg_names[i] for i in range(len(self.jit_fn.arg_names)) if i not in self.jit_fn.constexprs]

        self.constexpr_ids_set = set(self.constexpr_ids)

        self.cache = dict()

    def __getitem__(self, grid: Union[Tuple,Callable]):
        
        def wrapper(*args, **kwargs):

            nonlocal grid

            signature_list = list()

            # Get device ID
            if self.device_check:
                device_id = -1
                for arg in args:
                    if isinstance(arg, torch.Tensor):
                        device_id = arg.device.index
                        break
                
                if device_id == -1:
                    for k, v in kwargs.items():
                        if isinstance(v, torch.Tensor):
                            device_id = v.device.index
                            break

                signature_list.append(device_id)

            for i in self.constexpr_ids:
                if i >= len(args):
                    break
                signature_list.append(args[i])

            for k, v in kwargs.items():
                if k in self.constexpr_names:
                    signature_list.append((self.constexpr_names[k], v))

            if "batch_size" in kwargs:
                signature_list.append(("batch_size", kwargs["batch_size"]))

            if isinstance(grid, Callable):
                grid = grid(kwargs)

            grid_length = len(grid)
            grid0, grid1, grid2 = grid[0], grid[1] if grid_length > 1 else 1, grid[2] if grid_length > 2 else 1

            signature = tuple(signature_list)
            if signature in self.cache:
                kernel = self.cache[signature]

                aligned_args = list()
                for i, arg in enumerate(args):
                    if i not in self.constexpr_ids_set:
                        aligned_args.append(arg)

                for k in self.nonconstexpr_names:
                    if k in kwargs:
                        aligned_args.append(kwargs[k])

                if self.device_check:
                    with torch.cuda.device(device_id):
                        kernel[(grid0, grid1, grid2)](*aligned_args)
                else:
                    kernel[(grid0, grid1, grid2)](*aligned_args)
            else:
                if self.device_check:
                    with torch.cuda.device(device_id):
                        kernel = self.jit_fn[grid](*args, **kwargs)
                        self.cache[signature] = kernel
                else:
                    kernel = self.jit_fn[grid](*args, **kwargs)
                    self.cache[signature] = kernel

        return wrapper


def triton_jit(fn: Callable, device_check: bool = False):
    if torch.__version__[0] == "2" and torch.__version__[2] <= "2":
        return FastJITFunction(fn, device_check = device_check)
    else:
        return triton.jit(fn)



from __future__ import annotations

import torch
import pickle
from functools import partial
from typing import Sequence

from pyjuice.nodes import CircuitNodes, InputNodes, ProdNodes, SumNodes, inputs, multiply, summate


def serialize_nodes(root_ns: CircuitNodes):
    nodes_list = list()
    ns2id = dict()
    for ns in root_ns:
        if ns.is_input():
            ntype = "Input"
        elif ns.is_prod():
            ntype = "Product"
        else:
            assert ns.is_sum()
            ntype = "Sum"

        ns_info = {
            "type": ntype, 
            "num_node_blocks": ns.num_node_blocks,
            "block_size": ns.block_size,
            "chs": tuple(ns2id[cs] for cs in ns.chs)
        }

        if ns.is_prod() or ns.is_sum():
            ns_info["edge_ids"] = ns.edge_ids.detach().cpu().numpy().copy()

        if hasattr(ns, "_params") and ns._params is not None:
            ns_info["params"] = ns._params.detach().cpu().numpy().copy()

        if hasattr(ns, "_zero_param_mask") and ns._zero_param_mask is not None:
            ns_info["zero_param_mask"] = ns._zero_param_mask.detach().cpu().numpy().copy()

        if ns.is_input():
            ns_info["scope"] = ns.scope.to_list()
            ns_info["dist"] = pickle.dumps(ns.dist)

        ns2id[ns] = len(nodes_list)
        nodes_list.append(ns_info)

    for ns in root_ns:
        # Tied nodes
        if hasattr(ns, "_source_node") and ns._source_node is not None:
            nodes_list[ns2id[ns]]["source_node"] = ns2id[ns._source_node]

    return nodes_list


def deserialize_nodes(nodes_list: Sequence):
    id2ns = dict()
    for ns_id, ns_info in enumerate(nodes_list):
        if "num_node_blocks" in ns_info:
            num_node_blocks = ns_info["num_node_blocks"]
            block_size = ns_info["block_size"]
        elif "num_node_groups" in ns_info:
            # For backward compatibility. `num_node_blocks` was once called `num_node_groups`
            # and `block_size` was termed `group_size`
            num_node_blocks = ns_info["num_node_groups"]
            block_size = ns_info["group_size"]
        else:
            assert "num_nodes" in ns_info
            num_node_blocks = ns_info["num_nodes"]
            block_size = 1
        chids = ns_info["chs"]

        if ns_info["type"] == "Input":
            scope = ns_info["scope"]
            dist = pickle.loads(ns_info["dist"])

            ns = inputs(scope, num_node_blocks, dist, block_size = block_size, 
                        _no_set_meta_params = dist.need_meta_parameters)

            if "params" in ns_info:
                if dist.need_meta_parameters:
                    ns._params = torch.from_numpy(ns_info["params"])
                else:
                    ns.set_params(torch.from_numpy(ns_info["params"]), normalize = False)

        elif ns_info["type"] == "Product":
            chs = [id2ns[cid] for cid in chids]
            edge_ids = torch.from_numpy(ns_info["edge_ids"])

            if edge_ids.size(0) == chs[0].num_node_blocks:
                sparse_edges = False
            else:
                sparse_edges = True

            ns = multiply(*chs, edge_ids = edge_ids, sparse_edges = sparse_edges)

        else:
            assert ns_info["type"] == "Sum"

            chs = [id2ns[cid] for cid in chids]
            edge_ids = ns_info["edge_ids"]
            if "params" in ns_info:
                params = torch.from_numpy(ns_info["params"])
            else:
                params = None

            ns = summate(*chs, edge_ids = edge_ids, params = params, block_size = block_size)

            if "zero_param_mask" in ns_info:
                zero_param_mask = torch.from_numpy(ns_info["zero_param_mask"])
                ns.set_zero_param_mask(zero_param_mask)

        id2ns[ns_id] = ns

    for ns_id, ns_info in enumerate(nodes_list):
        if "source_node" in ns_info:
            ns = id2ns[ns_id]
            ns._source_node = id2ns[ns_info["source_node"]]

    return id2ns[len(nodes_list) - 1]



from __future__ import annotations

from functools import partial
from typing import Callable, Optional, Dict

from pyjuice.nodes import CircuitNodes, InputNodes, ProdNodes, SumNodes


def foreach(func: Callable, root_ns: CircuitNodes):
    """
    Traverse all nodes of a PC and can a specific function.

    :note: An alternative is to use the iterator directly via the `for` statement. See below for an example.

    :param func: the function to be called
    :type func: Callable

    :param root_ns: the root PC node
    :type root_ns: CircuitNodes

    Example::
        >>> for ns in root_ns: # Traverse the PC bottom-up
        ...     [do something...]
        >>> for ns in root_ns(reverse = True): # Traverse the PC top-down
        ...     [do something...]
    """

    visited = set()

    def dfs(ns: CircuitNodes):
        if ns in visited:
            return

        visited.add(ns)

        # Recursively traverse children
        if ns.is_sum() or ns.is_prod():
            for cs in ns.chs:
                dfs(cs)

        func(ns)

    dfs(root_ns)

    return None


def foldup_aggregate(func: Callable, root_ns: CircuitNodes, cache: Optional[Dict] = None, **kwargs):
    """
    Traverse all nodes of a PC bottom-up and aggregate a per-node object.

    :param func: the function to compute the per-node object
    :type func: Callable

    :param root_ns: the root PC node
    :type root_ns: CircuitNodes

    :param cache: an optional dictionary to store the per-node object for all nodes
    :type cache: Dict
    """

    if cache is None:
        cache = dict()

    if len(kwargs) > 0:
        func = partial(func, **kwargs)

    def dfs(ns: CircuitNodes):
        if ns in cache:
            return

        # Recursively traverse children
        if ns.is_sum() or ns.is_prod():
            for cs in ns.chs:
                dfs(cs)

        ch_outputs = [cache[cs] for cs in ns.chs]
        ns_output = func(ns, ch_outputs)

        cache[ns] = ns_output

    dfs(root_ns)

    return cache[root_ns]